#!/usr/bin/php -d open_basedir=/
<?php
/*
Copyright 2013-2016 Guillaume Boudreau

This file is part of Greyhole.

Greyhole is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Greyhole is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Greyhole.  If not, see <http://www.gnu.org/licenses/>.
*/

require_once('includes/common.php');
require_once('includes/CLI/CommandLineHelper.php'); // Command line helper (abstract classes, command line definitions & parsing, Runners, etc.)
require_once('includes/CLI/AbstractCliRunner.php');

class MoveCliRunner extends AbstractCliRunner {
    public function run() {
        $argc = $GLOBALS['argc'];
        $argv = $GLOBALS['argv'];

        if ($argc != 3) {
            die("Usage: greyhole-mv [source] [destination]\n  `source` and `destination` should start with a share name, following by the full path to a file or folder.\n  Example: greyhole-mv Videos/TV/24 VideosAttic/\n");
        }

        $source = $argv[1];
        $destination = $argv[2];

        foreach (Config::storagePoolDrives() as $sp_drive) {
            if (!StoragePool::is_pool_drive($sp_drive)) {
                continue;
            }
            if (file_exists("$sp_drive/$source")) {
                if (is_dir("$sp_drive/$source")) {
                    $full_destination = clean_dir("$sp_drive/$destination/" . basename($source));
                } else {
                    $full_destination = "$sp_drive/$destination";
                }

                gh_mv("$sp_drive/$source", "$sp_drive/$destination");
                echo("rm -rf $sp_drive/.gh_metastore/$source" . "\n");
                exec("rm -rf " . escapeshellarg("$sp_drive/.gh_metastore/$source"));
                if (file_exists("$sp_drive/.gh_metastore_backup/$source")) {
                    echo("rm -rf $sp_drive/.gh_metastore_backup/$source" . "\n");
                    exec("rm -rf " . escapeshellarg("$sp_drive/.gh_metastore_backup/$source"));
                }

                echo("greyhole --fsck --dir=" . escapeshellarg("$full_destination") . " --find-orphaned-files" . "\n");
                exec("greyhole --fsck --dir=" . escapeshellarg("$full_destination") . " --find-orphaned-files");
            }
        }
        $paths = explode('/', $source);
        $share = array_shift($paths);
        $full_path = implode('/', $paths);
        $landing_zone = get_share_landing_zone($share);

        echo("rm -rf " . escapeshellarg("$landing_zone/$full_path") . "\n");
        exec("rm -rf " . escapeshellarg("$landing_zone/$full_path"));

        echo("greyhole --fsck --dir=" . escapeshellarg(dirname("$landing_zone/$full_path")) . " --delete-orphaned-metadata" . "\n");
        exec("greyhole --fsck --dir=" . escapeshellarg(dirname("$landing_zone/$full_path")) . " --delete-orphaned-metadata");
    }
}

$GLOBALS['argv'][] = '--move';

// This will parse the command line, and find commands and options.
// It will then instantiate the appropriate class (that extends AbstractRunner), and return it.
$cliHelper = new CommandLineHelper();
$runner = $cliHelper->processCommandLine();

// This will run the runner(!) i.e. what the user asked for will now be executed.
$runner->run();

// The runner has finished; let's clean up before quitting.
$runner->finish();

function process_config() {
    if (!ConfigHelper::parse()) {
        Log::critical("Config file parsing failed. Exiting.", Log::EVENT_CODE_CONFIG_FILE_PARSING_FAILED);
    }
    // Config is OK; go on!
}

function gh_mv($source, $destination) {
	$orig_destination = $destination;
	if (is_dir($source)) {
		$destination = clean_dir($destination . '/' . basename($source));
	}
	$target_folder = dirname($destination);
	if (!file_exists($target_folder)) {
		echo("mkdir -p " . escapeshellarg($target_folder) . "\n");
		exec("mkdir -p " . escapeshellarg($target_folder));
	}
	if (!file_exists($destination)) {
		echo("mv " . escapeshellarg($source) . ' ' . escapeshellarg($orig_destination) . "\n");
		exec("mv " . escapeshellarg($source) . ' ' . escapeshellarg($orig_destination));
	} else {
		echo "Error: $destination already exists. Skipping renaming $source to $destination. You should fix this manually, then re-run greyhole-mv.\n";
	}
}

?>
