#!/usr/bin/php -d open_basedir=/
<?php
/*
Copyright 2013-2014 Guillaume Boudreau
This file is part of Greyhole.
Greyhole is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
Greyhole is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with Greyhole.  If not, see <http://www.gnu.org/licenses/>.
*/
define('CONFIG_LOG_LEVEL', 'log_level');
define('CONFIG_DELETE_MOVES_TO_TRASH', 'delete_moves_to_trash');
define('CONFIG_MODIFIED_MOVES_TO_TRASH', 'modified_moves_to_trash');
define('CONFIG_LOG_MEMORY_USAGE', 'log_memory_usage');
define('CONFIG_CHECK_FOR_OPEN_FILES', 'check_for_open_files');
define('CONFIG_ALLOW_MULTIPLE_SP_PER_DRIVE', 'allow_multiple_sp_per_device');
define('CONFIG_STORAGE_POOL_DRIVE', 'storage_pool_drive');
define('CONFIG_MIN_FREE_SPACE_POOL_DRIVE', 'min_free_space_pool_drive');
define('CONFIG_STICKY_FILES', 'sticky_files');
define('CONFIG_STICK_INTO', 'stick_into');
define('CONFIG_FROZEN_DIRECTORY', 'frozen_directory');
define('CONFIG_MEMORY_LIMIT', 'memory_limit');
define('CONFIG_TIMEZONE', 'timezone');
define('CONFIG_DRIVE_SELECTION_GROUPS', 'drive_selection_groups');
define('CONFIG_DRIVE_SELECTION_ALGORITHM', 'drive_selection_algorithm');
define('CONFIG_IGNORED_FILES', 'ignored_files');
define('CONFIG_IGNORED_FOLDERS', 'ignored_folders');
define('CONFIG_NUM_COPIES', 'num_copies');
define('CONFIG_LANDING_ZONE', 'landing_zone');
define('CONFIG_MAX_QUEUED_TASKS', 'max_queued_tasks');
define('CONFIG_EXECUTED_TASKS_RETENTION', 'executed_tasks_retention');
define('CONFIG_GREYHOLE_LOG_FILE', 'greyhole_log_file');
define('CONFIG_GREYHOLE_ERROR_LOG_FILE', 'greyhole_error_log_file');
define('CONFIG_EMAIL_TO', 'email_to');
define('CONFIG_DF_CACHE_TIME', 'df_cache_time');
define('CONFIG_DB_HOST', 'db_host');
define('CONFIG_DB_USER', 'db_user');
define('CONFIG_DB_PASS', 'db_pass');
define('CONFIG_DB_NAME', 'db_name');
define('CONFIG_METASTORE_BACKUPS', 'metastore_backups');
define('CONFIG_TRASH_SHARE', '===trash_share===');
define('CONFIG_HOOK', 'hook');
function recursive_include_parser($file) {
    $regex = '/^[ \t]*include[ \t]*=[ \t]*([^#\r\n]+)/im';
    $ok_to_execute = FALSE;
    if (is_array($file) && count($file) > 1) {
        $file = $file[1];
    }
    $file = trim($file);
    if (file_exists($file)) {
        if (is_executable($file)) {
            $perms = fileperms($file);
            $ok_to_execute = !($perms & 0x0080) || fileowner($file) === 0;
            $ok_to_execute &= !($perms & 0x0010) || filegroup($file) === 0;
            $ok_to_execute &= !($perms & 0x0002);
            if (!$ok_to_execute) {
                Log::warn("Config file '{$file}' is executable but file permissions are insecure, only the file's contents will be included.", Log::EVENT_CODE_CONFIG_INCLUDE_INSECURE_PERMISSIONS);
            }
        }
        $contents = $ok_to_execute ? shell_exec(escapeshellcmd($file)) : file_get_contents($file);
        return preg_replace_callback($regex, 'recursive_include_parser', $contents);
    } else {
        return false;
    }
}
final class ConfigHelper {
    static $df_command;
    public static $config_file = '/etc/greyhole.conf';
    public static $smb_config_file = '/etc/samba/smb.conf';
    public static $trash_share_names = array('Greyhole Attic', 'Greyhole Trash', 'Greyhole Recycle Bin');
    static $deprecated_options = array(
        'delete_moves_to_attic' => CONFIG_DELETE_MOVES_TO_TRASH,
        'storage_pool_directory' => CONFIG_STORAGE_POOL_DRIVE,
        'dir_selection_groups' => CONFIG_DRIVE_SELECTION_GROUPS,
        'dir_selection_algorithm' => CONFIG_DRIVE_SELECTION_ALGORITHM,
    );
    static $config_options = array(
        'bool' => array(
            CONFIG_DELETE_MOVES_TO_TRASH,
            CONFIG_MODIFIED_MOVES_TO_TRASH,
            CONFIG_LOG_MEMORY_USAGE,
            CONFIG_CHECK_FOR_OPEN_FILES,
            CONFIG_ALLOW_MULTIPLE_SP_PER_DRIVE,
        ),
        'number' => array(
            CONFIG_MAX_QUEUED_TASKS,
            CONFIG_EXECUTED_TASKS_RETENTION,
            CONFIG_DF_CACHE_TIME,
        ),
        'string' => array(
            CONFIG_DB_HOST,
            CONFIG_DB_USER,
            CONFIG_DB_PASS,
            CONFIG_DB_NAME,
            CONFIG_EMAIL_TO,
            CONFIG_GREYHOLE_LOG_FILE,
            CONFIG_GREYHOLE_ERROR_LOG_FILE,
            CONFIG_TIMEZONE,
            CONFIG_MEMORY_LIMIT,
        ),
    );
    public static function removeShare($share) {
        exec("/bin/sed -i 's/^.*num_copies\[".$share."\].*$//' " . escapeshellarg(self::$config_file));
    }
    public static function removeStoragePoolDrive($sp_drive) {
        $escaped_drive = str_replace('/', '\/', $sp_drive);
        exec("/bin/sed -i 's/^.*storage_pool_directory.*$escaped_drive.*$//' " . escapeshellarg(self::$config_file)); // Deprecated notation
        exec("/bin/sed -i 's/^.*storage_pool_drive.*$escaped_drive.*$//' " . escapeshellarg(self::$config_file));
    }
    public static function randomStoragePoolDrive() {
        $storage_pool_drives = (array) Config::storagePoolDrives();
        return $storage_pool_drives[array_rand($storage_pool_drives)];
    }
    public static function parse() {
        if (!ini_get('date.timezone')) {
            date_default_timezone_set('UTC');
        }
        $config_text = recursive_include_parser(self::$config_file);
        global $parsing_drive_selection_groups;
        foreach (explode("\n", $config_text) as $line) {
            if (preg_match("/^[ \t]*([^=\t]+)[ \t]*=[ \t]*([^#]+)/", $line, $regs)) {
                $name = trim($regs[1]);
                $value = trim($regs[2]);
                self::parse_line($name, $value);
            } else if ($parsing_drive_selection_groups !== FALSE) {
                $value = trim($line);
                if (strlen($value) == 0 || $value[0] == '#') {
                    continue;
                }
                if (preg_match("/(.+):(.+)/", $value, $regs)) {
                    $group_name = trim($regs[1]);
                    $drives = array_map('trim', explode(',', $regs[2]));
                    if (is_string($parsing_drive_selection_groups)) {
                        $share = $parsing_drive_selection_groups;
                        SharesConfig::add($share, CONFIG_DRIVE_SELECTION_GROUPS, $drives, $group_name);
                    } else {
                        Config::add(CONFIG_DRIVE_SELECTION_GROUPS, $drives, $group_name);
                    }
                }
            }
        }
        return self::init();
    }
    private static function parse_line($name, $value) {
        if ($name[0] == '#') {
            return;
        }
        self::normalize_name($name);
        global $parsing_drive_selection_groups;
        $parsing_drive_selection_groups = FALSE;
        if (self::parse_line_bool($name, $value)) return;
        if (self::parse_line_number($name, $value)) return;
        if (self::parse_line_string($name, $value)) return;
        if (self::parse_line_log($name, $value)) return;
        if (self::parse_line_pool_drive($name, $value)) return;
        if (self::parse_line_drive_selection($name, $value)) return;
        if (self::parse_line_sticky($name, $value)) return;
        if (self::parse_line_frozen($name, $value)) return;
        if (self::parse_line_ignore($name, $value)) return;
        if (self::parse_line_share_option($name, $value)) return;
        if (self::parse_line_hook($name, $value)) return;
        if (is_numeric($value)) {
            $value = (int) $value;
        }
        Config::set($name, $value);
    }
    private static function normalize_name(&$name) {
        foreach (self::$deprecated_options as $old_name => $new_name) {
            if (string_contains($name, $old_name)) {
                $fixed_name = str_replace($old_name, $new_name, $name);
                Log::warn("Deprecated option found in greyhole.conf: $name. You should change that to: $fixed_name", Log::EVENT_CODE_CONFIG_DEPRECATED_OPTION);
                $name = $fixed_name;
            }
        }
    }
    private static function parse_line_bool($name, $value) {
        if (array_contains(self::$config_options['bool'], $name)) {
            $bool = trim($value) === '1' || mb_stripos($value, 'yes') !== FALSE || mb_stripos($value, 'true') !== FALSE;
            Config::set($name, $bool);
            return TRUE;
        }
        return FALSE;
    }
    private static function parse_line_number($name, $value) {
        if (array_contains(self::$config_options['number'], $name)) {
            if (is_numeric($value)) {
                $value = (int) $value;
            }
            Config::set($name, $value);
            return TRUE;
        }
        return FALSE;
    }
    private static function parse_line_string($name, $value) {
        if (array_contains(self::$config_options['string'], $name)) {
            Config::set($name, $value);
            return TRUE;
        }
        return FALSE;
    }
    private static function parse_line_log($name, $value) {
        if ($name == CONFIG_LOG_LEVEL) {
            self::assert(defined("Log::$value"), "Invalid value for log_level: '$value'", Log::EVENT_CODE_CONFIG_INVALID_VALUE);
            Config::set(CONFIG_LOG_LEVEL, constant("Log::$value"));
            return TRUE;
        }
        return FALSE;
    }
    private static function parse_line_pool_drive($name, $value) {
        if ($name == CONFIG_STORAGE_POOL_DRIVE) {
            if (preg_match("/(.*) ?, ?min_free ?: ?([0-9]+) ?([gmk])b?/i", $value, $regs)) {
                $sp_drive = '/' . trim(trim($regs[1]), '/');
                Config::add(CONFIG_STORAGE_POOL_DRIVE, $sp_drive);
                $units = strtolower($regs[3]);
                if ($units == 'g') {
                    $value = (float) trim($regs[2]) * 1024.0 * 1024.0;
                } else if ($units == 'm') {
                    $value = (float) trim($regs[2]) * 1024.0;
                } else if ($units == 'k') {
                    $value = (float) trim($regs[2]);
                }
                Config::add(CONFIG_MIN_FREE_SPACE_POOL_DRIVE, $value, $sp_drive);
            } else {
                Log::warn("Warning! Unable to parse " . CONFIG_STORAGE_POOL_DRIVE . " line from config file. Value = $value", Log::EVENT_CODE_CONFIG_UNPARSEABLE_LINE);
            }
            return TRUE;
        }
        return FALSE;
    }
    private static function parse_line_sticky($name, $value) {
        if ($name == CONFIG_STICKY_FILES) {
            global $last_sticky_files_dir;
            $last_sticky_files_dir = trim($value, '/');
            Config::add(CONFIG_STICKY_FILES, array(), $last_sticky_files_dir);
            return TRUE;
        }
        if ($name == CONFIG_STICK_INTO) {
            global $last_sticky_files_dir;
            $sticky_files = Config::get(CONFIG_STICKY_FILES);
            $sticky_files[$last_sticky_files_dir][] = '/' . trim($value, '/');
            Config::set(CONFIG_STICKY_FILES, $sticky_files);
            return TRUE;
        }
        return FALSE;
    }
    private static function parse_line_frozen($name, $value) {
        if ($name == CONFIG_FROZEN_DIRECTORY) {
            Config::add(CONFIG_FROZEN_DIRECTORY, trim($value, '/'));
            return TRUE;
        }
        return FALSE;
    }
    private static function parse_line_drive_selection($name, $value) {
        if ($name == CONFIG_DRIVE_SELECTION_GROUPS) {
            if (preg_match("/(.+):(.*)/", $value, $regs)) {
                $group_name = trim($regs[1]);
                $group_definition = array_map('trim', explode(',', $regs[2]));
                Config::add(CONFIG_DRIVE_SELECTION_GROUPS, $group_definition, $group_name);
                global $parsing_drive_selection_groups;
                $parsing_drive_selection_groups = TRUE;
            }
            return TRUE;
        }
        if ($name == CONFIG_DRIVE_SELECTION_ALGORITHM) {
            Config::set(CONFIG_DRIVE_SELECTION_ALGORITHM, PoolDriveSelector::parse($value, Config::get(CONFIG_DRIVE_SELECTION_GROUPS)));
            return TRUE;
        }
        return FALSE;
    }
    private static function parse_line_ignore($name, $value) {
        if ($name == CONFIG_IGNORED_FILES) {
            Config::add(CONFIG_IGNORED_FILES, $value);
            return TRUE;
        }
        if ($name == CONFIG_IGNORED_FOLDERS) {
            Config::add(CONFIG_IGNORED_FOLDERS, $value);
            return TRUE;
        }
        return FALSE;
    }
    private static function parse_line_share_option($name, $value) {
        if (string_starts_with($name, CONFIG_NUM_COPIES)) {
            $share = mb_substr($name, 11, mb_strlen($name)-12);
            if (mb_stripos($value, 'max') === 0) {
                $value = 9999;
            }
            SharesConfig::set($share, CONFIG_NUM_COPIES, (int) $value);
        } else if (string_starts_with($name, CONFIG_DELETE_MOVES_TO_TRASH)) {
            $share = mb_substr($name, 22, mb_strlen($name)-23);
            $value = strtolower(trim($value));
            $bool = trim($value) === '1' || mb_stripos($value, 'yes') !== FALSE || mb_stripos($value, 'true') !== FALSE;
            SharesConfig::set($share, CONFIG_DELETE_MOVES_TO_TRASH, $bool);
        } else if (string_starts_with($name, CONFIG_MODIFIED_MOVES_TO_TRASH)) {
            $share = mb_substr($name, 24, mb_strlen($name)-25);
            $value = strtolower(trim($value));
            $bool = trim($value) === '1' || mb_stripos($value, 'yes') !== FALSE || mb_stripos($value, 'true') !== FALSE;
            SharesConfig::set($share, CONFIG_MODIFIED_MOVES_TO_TRASH, $bool);
        } else if (string_starts_with($name, CONFIG_DRIVE_SELECTION_GROUPS)) {
            $share = mb_substr($name, 23, mb_strlen($name)-24);
            if (preg_match("/(.+):(.+)/", $value, $regs)) {
                $group_name = trim($regs[1]);
                $group_definition = array_map('trim', explode(',', $regs[2]));
                SharesConfig::add($share, CONFIG_DRIVE_SELECTION_GROUPS, $group_definition, $group_name);
                global $parsing_drive_selection_groups;
                $parsing_drive_selection_groups = $share;
            }
        } else if (string_starts_with($name, CONFIG_DRIVE_SELECTION_ALGORITHM)) {
            $share = mb_substr($name, 26, mb_strlen($name)-27);
            if (SharesConfig::get($share, CONFIG_DRIVE_SELECTION_GROUPS) === FALSE) {
                SharesConfig::set($share, CONFIG_DRIVE_SELECTION_GROUPS, Config::get(CONFIG_DRIVE_SELECTION_GROUPS));
            }
            SharesConfig::set($share, CONFIG_DRIVE_SELECTION_ALGORITHM, PoolDriveSelector::parse($value, SharesConfig::get($share, CONFIG_DRIVE_SELECTION_GROUPS)));
        } else {
            return FALSE;
        }
        return TRUE;
    }
    private static function parse_line_hook($name, $value) {
        if (string_starts_with($name, CONFIG_HOOK)) {
            if (!preg_match('/hook\[([^\]]+)\]/', $name, $re)) {
                Log::warn("Can't parse the following config line: $name; ignoring.", Log::EVENT_CODE_CONFIG_UNPARSEABLE_LINE);
                return TRUE;
            }
            if (!is_executable($value)) {
                Log::warn("Hook script $value is not executable; ignoring.", Log::EVENT_CODE_CONFIG_HOOK_SCRIPT_NOT_EXECUTABLE);
                return TRUE;
            }
            $events = explode('|', $re[1]);
            foreach ($events as $event) {
                Hook::add($event, $value);
            }
            return TRUE;
        }
        return FALSE;
    }
    private static function init() {
        Log::setLevel(Config::get(CONFIG_LOG_LEVEL));
        if (count(Config::storagePoolDrives()) == 0) {
            Log::error("You have no '" . CONFIG_STORAGE_POOL_DRIVE . "' defined. Greyhole can't run.", Log::EVENT_CODE_CONFIG_NO_STORAGE_POOL);
            return FALSE;
        }
        self::$df_command = "df -k";
        foreach (Config::storagePoolDrives() as $sp_drive) {
            self::$df_command .= " " . escapeshellarg($sp_drive);
        }
        self::$df_command .= " 2>&1 | grep '%' | grep -v \"^df: .*: No such file or directory$\"";
        exec('testparm -s ' . escapeshellarg(self::$smb_config_file) . ' 2> /dev/null', $config_text);
        foreach ($config_text as $line) {
            $line = trim($line);
            if (mb_strlen($line) == 0) { continue; }
            if ($line[0] == '[' && preg_match('/\[([^\]]+)\]/', $line, $regs)) {
                $share_name = $regs[1];
            }
            if (isset($share_name) && !SharesConfig::exists($share_name) && !array_contains(self::$trash_share_names, $share_name)) {
                continue;
            }
            if (isset($share_name) && preg_match('/^\s*path[ \t]*=[ \t]*(.+)$/i', $line, $regs)) {
                SharesConfig::set($share_name, CONFIG_LANDING_ZONE, '/' . trim($regs[1], '/"'));
                SharesConfig::set($share_name, 'name', $share_name);
            }
        }
        $drive_selection_algorithm = Config::get(CONFIG_DRIVE_SELECTION_ALGORITHM);
        if (!empty($drive_selection_algorithm)) {
            foreach ($drive_selection_algorithm as $ds) {
                $ds->update();
            }
        } else {
            $drive_selection_algorithm = PoolDriveSelector::parse('most_available_space', null);
        }
        Config::set(CONFIG_DRIVE_SELECTION_ALGORITHM, $drive_selection_algorithm);
        if (!Config::exists(CONFIG_MODIFIED_MOVES_TO_TRASH)) {
            Config::set(CONFIG_MODIFIED_MOVES_TO_TRASH, Config::get(CONFIG_DELETE_MOVES_TO_TRASH));
        }
        foreach (SharesConfig::getShares() as $share_name => $share_options) {
            if (array_contains(self::$trash_share_names, $share_name)) {
                SharesConfig::set(CONFIG_TRASH_SHARE, 'name', $share_name);
                SharesConfig::set(CONFIG_TRASH_SHARE, CONFIG_LANDING_ZONE, SharesConfig::get($share_name, CONFIG_LANDING_ZONE));
                SharesConfig::removeShare($share_name);
                continue;
            }
            if ($share_options[CONFIG_NUM_COPIES] > count(Config::storagePoolDrives())) {
                SharesConfig::set($share_name, CONFIG_NUM_COPIES, count(Config::storagePoolDrives()));
            }
            if (!isset($share_options[CONFIG_LANDING_ZONE])) {
                Log::warn("Found a share ($share_name) defined in " . self::$config_file . " with no path in " . self::$smb_config_file . ". Either add this share in " . self::$smb_config_file . ", or remove it from " . self::$config_file . ", then restart Greyhole.", Log::EVENT_CODE_CONFIG_SHARE_MISSING_FROM_SMB_CONF);
                return FALSE;
            }
            if (!isset($share_options[CONFIG_DELETE_MOVES_TO_TRASH])) {
                SharesConfig::set($share_name, CONFIG_DELETE_MOVES_TO_TRASH, Config::get(CONFIG_DELETE_MOVES_TO_TRASH));
            }
            if (!isset($share_options[CONFIG_MODIFIED_MOVES_TO_TRASH])) {
                SharesConfig::set($share_name, CONFIG_MODIFIED_MOVES_TO_TRASH, SharesConfig::get($share_name, CONFIG_DELETE_MOVES_TO_TRASH));
            }
            if (isset($share_options[CONFIG_DRIVE_SELECTION_ALGORITHM])) {
                foreach ($share_options[CONFIG_DRIVE_SELECTION_ALGORITHM] as $ds) {
                    $ds->update();
                }
            } else {
                SharesConfig::set($share_name, CONFIG_DRIVE_SELECTION_ALGORITHM, $drive_selection_algorithm);
            }
            if (isset($share_options[CONFIG_DRIVE_SELECTION_GROUPS])) {
                SharesConfig::remove($share_name, CONFIG_DRIVE_SELECTION_GROUPS);
            }
            foreach (Config::storagePoolDrives() as $sp_drive) {
                if (string_starts_with($share_options[CONFIG_LANDING_ZONE], $sp_drive)) {
                    Log::critical("Found a share ($share_name), with path " . $share_options[CONFIG_LANDING_ZONE] . ", which is INSIDE a storage pool drive ($sp_drive). Share directories should never be inside a directory that you have in your storage pool.\nFor your shares to use your storage pool, you just need them to have 'vfs objects = greyhole' in their (smb.conf) config; their location on your file system is irrelevant.", Log::EVENT_CODE_CONFIG_LZ_INSIDE_STORAGE_POOL);
                }
                if (string_starts_with($sp_drive, $share_options[CONFIG_LANDING_ZONE])) {
                    Log::critical("Found a storage pool drive ($sp_drive), which is INSIDE a share landing zone (" . $share_options[CONFIG_LANDING_ZONE] . "), for share $share_name. Storage pool drives should never be inside a directory that you use as a share landing zone ('path' in smb.conf).\nFor your shares to use your storage pool, you just need them to have 'vfs objects = greyhole' in their (smb.conf) config; their location on your file system is irrelevant.", Log::EVENT_CODE_CONFIG_STORAGE_POOL_INSIDE_LZ);
                }
            }
        }
        foreach (Config::storagePoolDrives() as $sp_drive) {
            $found = FALSE;
            foreach (SharesConfig::getShares() as $share_name => $share_options) {
                foreach ($share_options[CONFIG_DRIVE_SELECTION_ALGORITHM] as $ds) {
                    if (array_contains($ds->drives, $sp_drive)) {
                        $found = TRUE;
                    }
                }
            }
            if (!$found) {
                Log::warn("The storage pool drive '$sp_drive' is not part of any drive_selection_algorithm definition, and will thus never be used to receive any files.", Log::EVENT_CODE_CONFIG_STORAGE_POOL_DRIVE_NOT_IN_DRIVE_SELECTION_ALGO);
            }
        }
        $memory_limit = Config::get(CONFIG_MEMORY_LIMIT);
        ini_set('memory_limit', $memory_limit);
        if (preg_match('/G$/i',$memory_limit)) {
            $memory_limit = preg_replace('/G$/i','',$memory_limit);
            $memory_limit = $memory_limit * 1024 * 1024 * 1024;
        } else if (preg_match('/M$/i',$memory_limit)) {
            $memory_limit = preg_replace('/M$/i','',$memory_limit);
            $memory_limit = $memory_limit * 1024 * 1024;
        } else if (preg_match('/K$/i',$memory_limit)) {
            $memory_limit = preg_replace('/K$/i','',$memory_limit);
            $memory_limit = $memory_limit * 1024;
        }
        Config::set(CONFIG_MEMORY_LIMIT, $memory_limit);
        $tz = Config::get(CONFIG_TIMEZONE);
        if (empty($tz)) {
            $tz = @date_default_timezone_get();
        }
        date_default_timezone_set($tz);
        $db_options = array(
            'engine' => 'mysql',
            'schema' => "/usr/share/greyhole/schema-mysql.sql",
            'host' => Config::get(CONFIG_DB_HOST),
            'user' => Config::get(CONFIG_DB_USER),
            'pass' => Config::get(CONFIG_DB_PASS),
            'name' => Config::get(CONFIG_DB_NAME),
        );
        DB::setOptions($db_options);
        if (strtolower(Config::get(CONFIG_GREYHOLE_LOG_FILE)) == 'syslog') {
            openlog("Greyhole", LOG_PID, LOG_USER);
        }
        return TRUE;
    }
    private static function assert($check, $error_message, $event_code) {
        if ($check === FALSE) {
            Log::critical($error_message, $event_code);
        }
    }
}
final class Config {
    public static $config = array(
        CONFIG_LOG_LEVEL                   => Log::DEBUG,
        CONFIG_DELETE_MOVES_TO_TRASH       => TRUE,
        CONFIG_LOG_MEMORY_USAGE            => FALSE,
        CONFIG_CHECK_FOR_OPEN_FILES        => TRUE,
        CONFIG_ALLOW_MULTIPLE_SP_PER_DRIVE => FALSE,
        CONFIG_STORAGE_POOL_DRIVE          => array(),
        CONFIG_MIN_FREE_SPACE_POOL_DRIVE   => array(),
        CONFIG_STICKY_FILES                => array(),
        CONFIG_FROZEN_DIRECTORY            => array(),
        CONFIG_MEMORY_LIMIT                => '512M',
        CONFIG_TIMEZONE                    => FALSE,
        CONFIG_DRIVE_SELECTION_GROUPS      => array(),
        CONFIG_IGNORED_FILES               => array(),
        CONFIG_IGNORED_FOLDERS             => array(),
        CONFIG_MAX_QUEUED_TASKS            => 10000000,
        CONFIG_EXECUTED_TASKS_RETENTION    => 60,
        CONFIG_GREYHOLE_LOG_FILE           => '/var/log/greyhole.log',
        CONFIG_GREYHOLE_ERROR_LOG_FILE     => FALSE,
        CONFIG_EMAIL_TO                    => 'root',
        CONFIG_DF_CACHE_TIME               => 15,
    );
    public static function get($name, $index=NULL) {
        if ($index === NULL) {
            return isset(self::$config[$name]) ? self::$config[$name] : FALSE;
        } else {
            return isset(self::$config[$name][$index]) ? self::$config[$name][$index] : FALSE;
        }
    }
    public static function exists($name) {
        return isset(self::$config[$name]);
    }
    public static function storagePoolDrives() {
        return self::get(CONFIG_STORAGE_POOL_DRIVE);
    }
    public static function set($name, $value) {
        self::$config[$name] = $value;
    }
    public static function add($name, $value, $index=NULL) {
        if ($index === NULL) {
            self::$config[$name][] = $value;
        } else {
            self::$config[$name][$index] = $value;
        }
    }
}
final class SharesConfig {
    private static $shares_config;
    private static function _getConfig($share) {
        if (!self::exists($share)) {
            self::$shares_config[$share] = array();
        }
        return self::$shares_config[$share];
    }
    public static function exists($share) {
        return isset(self::$shares_config[$share]);
    }
    public static function getShares() {
        $result = array();
        foreach (self::$shares_config as $share_name => $share_config) {
            if ($share_name != CONFIG_TRASH_SHARE) {
                $result[$share_name] = $share_config;
            }
        }
        return $result;
    }
    public static function getConfigForShare($share) {
        if (!self::exists($share)) {
            return FALSE;
        }
        return self::$shares_config[$share];
    }
    public static function removeShare($share) {
        unset(self::$shares_config[$share]);
    }
    public static function remove($share, $name) {
        unset(self::$shares_config[$share][$name]);
    }
    public static function get($share, $name, $index=NULL) {
        if (!self::exists($share)) {
            return FALSE;
        }
        $config = self::$shares_config[$share];
        if ($index === NULL) {
            return isset($config[$name]) ? $config[$name] : FALSE;
        } else {
            return isset($config[$name][$index]) ? $config[$name][$index] : FALSE;
        }
    }
    public static function set($share, $name, $value) {
        $config = self::_getConfig($share);
        $config[$name] = $value;
        self::$shares_config[$share] = $config;
    }
    public static function add($share, $name, $value, $index=NULL) {
        $config = self::_getConfig($share);
        if ($index === NULL) {
            $config[$name][] = $value;
        } else {
            $config[$name][$index] = $value;
        }
        self::$shares_config[$share] = $config;
    }
}

final class DB {
	protected static $options; // connection options
	protected static $handle; // database handle
	public static function setOptions($options) {
        if (is_array($options)) {
            $options = (object) $options;
        }
		self::$options = $options;
	}
	public static function connect($retry_until_successful=FALSE) {
        $connect_string = 'mysql:host=' . self::$options->host . ';dbname=' . self::$options->name;
        try {
            self::$handle = @new PDO($connect_string, self::$options->user, self::$options->pass, array(PDO::ATTR_TIMEOUT => 10, PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION));
        } catch (PDOException $ex) {
            if ($retry_until_successful) {
                sleep(2);
                return DB::connect(TRUE);
            }
            echo "ERROR: Can't connect to database: " . $ex->getMessage() . "\n";
            Log::critical("Can't connect to database: " . $ex->getMessage(), Log::EVENT_CODE_DB_CONNECT_FAILED);
        }
        if (self::$handle) {
            DB::execute("SET SESSION group_concat_max_len = 1048576");
            DB::execute("SET SESSION wait_timeout = 86400"); # Allow 24h fsck!
            DB::migrate();
        }
        return self::$handle;
    }
    public static function execute($q, $args = array(), $attempt_repair=TRUE) {
        $stmt = self::$handle->prepare($q);
        foreach ($args as $key => $value) {
            $stmt->bindValue(":$key", $value);
        }
        try {
            $stmt->execute();
            return $stmt;
        } catch (PDOException $e) {
            $error = self::$handle->errorInfo();
            if (($error[1] == 144 || $error[1] == 145) && $attempt_repair) {
                Log::info("Error during MySQL query: " . $e->getMessage() . '. Will now try to repair the MySQL tables.');
                DB::repairTables();
                return DB::execute($q, $args, FALSE); // $attempt_repair = FALSE, to not go into an infinite loop, if the repair doesn't work.
            }
            throw new Exception($e->getMessage(), $error[1]);
        }
    }
    public static function insert($q, $args = array()) {
        if (DB::execute($q, $args) === FALSE) {
            return FALSE;
        }
        return DB::lastInsertedId();
    }
    public static function getFirst($q, $args = array()) {
        $stmt = DB::execute($q, $args);
        $result = $stmt->fetch(PDO::FETCH_ASSOC);
        if ($result === FALSE) {
            return FALSE;
        }
        return (object) $result;
    }
    public static function getFirstValue($q, $args = array()) {
        $row = (array) DB::getFirst($q, $args);
        if (!is_array($row)) {
            return FALSE;
        }
        return array_shift($row);
    }
    public static function getAll($q, $args = array(), $index_field=NULL) {
        $stmt = DB::execute($q, $args);
        $rows = array();
        $i = 0;
        while ($row = $stmt->fetch(PDO::FETCH_ASSOC)) {
            $index = $i++;
            if (!empty($index_field)) {
                $index = $row[$index_field];
            }
            $rows[$index] = (object) $row;
        }
        return $rows;
    }
    public static function getAllValues($q, $args = array(), $data_type=null) {
        $stmt = DB::execute($q, $args);
        $values = array();
        while ($row = $stmt->fetch(PDO::FETCH_ASSOC)) {
            if (!is_array($row)) {
                return FALSE;
            }
            $value = array_shift($row);
            if (!empty($data_type)) {
                settype($value, $data_type);
            }
            $values[] = $value;
        }
        return $values;
    }
    public static function lastInsertedId() {
        $q = "SELECT LAST_INSERT_ID()";
        $lastInsertedId = (int) DB::getFirstValue($q);
        if ($lastInsertedId === 0) {
            return TRUE;
        }
        return $lastInsertedId;
    }
    public static function quote($string) {
        $escaped_string = self::$handle->quote($string);
        return substr($escaped_string, 1, -1);
    }
    public static function error() {
        return self::$options->error;
    }
    private static function migrate() {
        DB::migrate_1_frozen_thaw();
        DB::migrate_2_idle();
        DB::migrate_3_larger_settings();
        DB::migrate_4_find_next_task_index();
        DB::migrate_5_find_next_task_index();
        DB::migrate_6_md5_worker_indexes();
        DB::migrate_7_larger_full_path();
        DB::migrate_8_du_stats();
        DB::migrate_9_complete_writen();
        DB::migrate_10_utf8();
    }
    private static function migrate_1_frozen_thaw() {
        $query = "DESCRIBE tasks";
        $rows = DB::getAll($query);
        foreach ($rows as $row) {
            if ($row->Field == 'complete') {
                if ($row->Type == "enum('yes','no')") {
                    DB::execute("ALTER TABLE tasks CHANGE complete complete ENUM('yes','no','frozen','thawed') NOT NULL");
                    DB::execute("ALTER TABLE tasks_completed CHANGE complete complete ENUM('yes','no','frozen','thawed') NOT NULL");
                }
                break;
            }
        }
    }
    private static function migrate_2_idle() {
        $query = "DESCRIBE tasks";
        $rows = DB::getAll($query);
        foreach ($rows as $row) {
            if ($row->Field == 'complete') {
                if ($row->Type == "enum('yes','no','frozen','thawed')") {
                    DB::execute("ALTER TABLE tasks CHANGE complete complete ENUM('yes','no','frozen','thawed','idle') NOT NULL");
                    DB::execute("ALTER TABLE tasks_completed CHANGE complete complete ENUM('yes','no','frozen','thawed','idle') NOT NULL");
                }
                break;
            }
        }
    }
    private static function migrate_3_larger_settings() {
        $query = "DESCRIBE settings";
        $rows = DB::getAll($query);
        foreach ($rows as $row) {
            if ($row->Field == 'value') {
                if ($row->Type == "tinytext") {
                    DB::execute("ALTER TABLE settings CHANGE value value TEXT CHARACTER SET utf8 NOT NULL");
                }
                break;
            }
        }
    }
    private static function migrate_4_find_next_task_index() {
        $query = "SHOW INDEX FROM tasks WHERE Key_name = 'find_next_task'";
        $row = DB::getFirst($query);
        if ($row === FALSE) {
            DB::execute("ALTER TABLE tasks ADD INDEX find_next_task (complete, share(64), id)");
        }
        $query = "SHOW INDEX FROM tasks WHERE Key_name = 'incomplete_open'";
        $row = DB::getFirst($query);
        if ($row) {
            DB::execute("ALTER TABLE tasks DROP INDEX incomplete_open");
        }
        $query = "SHOW INDEX FROM tasks WHERE Key_name = 'subsequent_writes'";
        $row = DB::getFirst($query);
        if ($row) {
            DB::execute("ALTER TABLE tasks DROP INDEX subsequent_writes");
        }
        $query = "SHOW INDEX FROM tasks WHERE Key_name = 'unneeded_unlinks'";
        $row = DB::getFirst($query);
        if ($row) {
            DB::execute("ALTER TABLE tasks DROP INDEX unneeded_unlinks");
        }
    }
    private static function migrate_5_find_next_task_index() {
        $query = "SHOW INDEX FROM tasks WHERE Key_name = 'find_next_task' and Column_name = 'share'";
        $row = DB::getFirst($query);
        if ($row !== FALSE) {
            DB::execute("ALTER TABLE tasks DROP INDEX find_next_task");
            DB::execute("ALTER TABLE tasks ADD INDEX find_next_task (complete, id)");
        }
    }
    private static function migrate_6_md5_worker_indexes() {
        $query = "SHOW INDEX FROM tasks WHERE Key_name = 'md5_worker'";
        $row = DB::getFirst($query);
        if ($row === FALSE) {
            DB::execute("ALTER TABLE tasks ADD INDEX md5_worker (action, complete, additional_info(100), id)");
        }
        $query = "SHOW INDEX FROM tasks WHERE Key_name = 'md5_checker'";
        $row = DB::getFirst($query);
        if ($row === FALSE) {
            DB::execute("ALTER TABLE tasks ADD INDEX md5_checker (action, share(64), full_path(265), complete)");
        }
        $query = "DESCRIBE tasks";
        $rows = DB::getAll($query);
        foreach ($rows as $row) {
            if ($row->Field == 'additional_info') {
                if ($row->Type == "tinytext") {
                    DB::execute("ALTER TABLE tasks CHANGE additional_info additional_info TEXT CHARACTER SET utf8 NULL");
                }
                break;
            }
        }
    }
    private static function migrate_7_larger_full_path() {
        $query = "DESCRIBE tasks";
        $rows = DB::getAll($query);
        foreach ($rows as $row) {
            if ($row->Field == 'full_path') {
                if ($row->Type == "tinytext") {
                    DB::execute("ALTER TABLE tasks CHANGE full_path full_path TEXT CHARACTER SET utf8 NULL");
                    DB::execute("ALTER TABLE tasks_completed CHANGE full_path full_path TEXT CHARACTER SET utf8 NULL");
                }
                break;
            }
        }
    }
    private static function migrate_8_du_stats() {
        $query = "CREATE TABLE IF NOT EXISTS `du_stats` (`share` TINYTEXT NOT NULL, `full_path` TEXT NOT NULL, `depth` TINYINT(3) UNSIGNED NOT NULL, `size` BIGINT(20) UNSIGNED NOT NULL, UNIQUE KEY `uniqness` (`share`(64),`full_path`(269))) ENGINE = MYISAM DEFAULT CHARSET=utf8";
        DB::execute($query);
        $query = "SHOW INDEX FROM `du_stats` WHERE Key_name = 'uniqness'";
        $row = DB::getFirst($query);
        if ($row === FALSE) {
            DB::execute("TRUNCATE `du_stats`");
            DB::execute("ALTER TABLE `du_stats` ADD UNIQUE `uniqness` (`share` (64), `full_path` (269))");
        }
    }
    private static function migrate_9_complete_writen() {
        $query = "DESCRIBE tasks";
        $rows = DB::getAll($query);
        foreach ($rows as $row) {
            if ($row->Field == 'complete') {
                if ($row->Type == "enum('yes','no','frozen','thawed','idle')") {
                    DB::execute("ALTER TABLE tasks CHANGE complete complete ENUM('yes','no','frozen','thawed','idle','written') NOT NULL");
                    DB::execute("ALTER TABLE tasks_completed CHANGE complete complete ENUM('yes','no','frozen','thawed','idle','written') NOT NULL");
                }
                break;
            }
        }
    }
    private static function migrate_10_utf8() {
        $q = "SHOW INDEX FROM du_stats WHERE key_name = 'uniqness' AND column_name = 'full_path'";
        $index_def = DB::getFirst($q);
        if ($index_def->Sub_part > 269) {
            $q = "ALTER TABLE du_stats DROP INDEX uniqness";
            DB::execute($q);
            $q = "ALTER TABLE du_stats ADD UNIQUE INDEX `uniqness` (share(64), full_path(269))";
            DB::execute($q);
        }
        $q = "SHOW INDEX FROM tasks WHERE key_name = 'md5_checker' AND column_name = 'full_path'";
        $index_def = DB::getFirst($q);
        if ($index_def->Sub_part > 265) {
            $q = "ALTER TABLE tasks DROP INDEX md5_checker";
            DB::execute($q);
            $q = "ALTER TABLE tasks ADD INDEX `md5_checker` (action, share(64), full_path(265), complete)";
            DB::execute($q);
        }
        $tables = array(
            'du_stats',
            'settings',
            'tasks',
            'tasks_completed'
        );
        $columns = array(
            'du_stats|share|TINYTEXT CHARACTER SET utf8 NOT NULL',
            'du_stats|full_path|TEXT CHARACTER SET utf8 NOT NULL',
            'settings|name|TINYTEXT CHARACTER SET utf8 NOT NULL',
            'settings|value|TEXT CHARACTER SET utf8 NOT NULL',
            'tasks|share|TINYTEXT CHARACTER SET utf8 NOT NULL',
            'tasks|full_path|TEXT CHARACTER SET utf8 NULL',
            'tasks|additional_info|TEXT CHARACTER SET utf8 NULL',
            'tasks_completed|share|TINYTEXT CHARACTER SET utf8 NOT NULL',
            'tasks_completed|full_path|TEXT CHARACTER SET utf8 NULL',
            'tasks_completed|additional_info|TEXT CHARACTER SET utf8 NULL',
        );
        $query = "SELECT CCSA.character_set_name FROM information_schema.`TABLES` T, information_schema.`COLLATION_CHARACTER_SET_APPLICABILITY` CCSA WHERE CCSA.collation_name = T.table_collation AND T.table_schema = :schema AND T.table_name = :table";
        foreach ($tables as $table_name) {
            $charset = DB::getFirstValue($query, array('schema' => Config::get(CONFIG_DB_NAME), 'table' => $table_name));
            if ($charset != "utf8") {
                Log::info("Updating $table_name table to UTF-8");
                try {
                    DB::execute("ALTER TABLE `$table_name` CONVERT TO CHARACTER SET utf8 COLLATE utf8_general_ci");
                } catch (Exception $ex) {
                    try {
                        DB::execute("ALTER TABLE `$table_name` CHARACTER SET utf8 COLLATE utf8_general_ci");
                    } catch (Exception $ex) {
                        Log::warn("  ALTER TABLE failed.", Log::EVENT_CODE_DB_MIGRATION_FAILED);
                    }
                }
            }
        }
        $query = "SELECT character_set_name FROM information_schema.`COLUMNS` C WHERE table_schema = :schema AND table_name = :table AND column_name = :field";
        foreach ($columns as $value) {
            list($table_name, $column_name, $definition) = explode('|', $value);
            $charset = DB::getFirstValue($query, array('schema' => Config::get(CONFIG_DB_NAME), 'table' => $table_name, 'field' => $column_name));
            if ($charset != "utf8") {
                Log::info("Updating $table_name.$column_name column to UTF-8");
                DB::execute("ALTER TABLE `$table_name` CHANGE `$column_name` `$column_name` $definition");
            }
        }
    }
    public static function repairTables() {
        if (Log::actionIs(ACTION_DAEMON)) {
            Log::info("Optimizing MySQL tables...");
        }
        foreach (array('tasks', 'settings', 'du_stats', 'tasks_completed') as $table_name) {
            try {
                DB::getFirst("SELECT * FROM $table_name LIMIT 1");
            } catch (Exception $e) {
                Log::info("Repairing $table_name MySQL table...");
                DB::execute("REPAIR TABLE $table_name", array(), FALSE);
            }
        }
    }
    public static function deleteExecutedTasks() {
        $executed_tasks_retention = Config::get(CONFIG_EXECUTED_TASKS_RETENTION);
        if ($executed_tasks_retention == 'forever') {
            return;
        }
        if (!is_int($executed_tasks_retention)) {
            Log::critical("Error: Invalid value for 'executed_tasks_retention' in greyhole.conf: '$executed_tasks_retention'. You need to use either 'forever' (no quotes), or a number of days.", Log::EVENT_CODE_CONFIG_INVALID_VALUE);
        }
        Log::info("Cleaning executed tasks: keeping the last $executed_tasks_retention days of logs.");
        $query = sprintf("DELETE FROM tasks_completed WHERE event_date < NOW() - INTERVAL %d DAY", (int) $executed_tasks_retention);
        DB::execute($query);
    }
}

abstract class Hook
{
    protected $event_type;
    protected $script;
    protected static $hooks = array();
    public function __construct($event_type, $script) {
        $this->event_type = $event_type;
        $this->script = $script;
    }
    public static function add($event_type, $script) {
        if (array_contains(FileHook::getEventTypes(), $event_type)) {
            $hook = new FileHook($event_type, $script);
        } elseif (array_contains(LogHook::getEventTypes(), $event_type)) {
            $hook = new LogHook($event_type, $script);
        } else {
            Log::warn("Unknown hook event type '$event_type'; ignoring.", Log::EVENT_CODE_HOOK_NOT_EXECUTABLE);
            return;
        }
        static::$hooks[$event_type][] = $hook;
    }
    public static function hasHookForEvent($event_type) {
        return !empty(static::$hooks[$event_type]);
    }
    protected static function _trigger($event_type, $context) {
        $hooks = @static::$hooks[$event_type];
        if (!is_array($hooks)) {
            return;
        }
        foreach ($hooks as $hook) {
            Log::debug("Calling external hook $hook->script for event $hook->event_type ...");
            exec(escapeshellarg($hook->script) . " " . implode(' ', $hook->getArgs($context)) . " 2>&1", $output, $result_code);
            if (!empty($output)) {
                foreach($output as $line) {
                    Log::debug("  $line");
                }
            }
            if ($result_code === 0) {
                Log::debug("External hook exited with status code $result_code.");
            } else {
                if ($hook->event_type == LogHook::EVENT_TYPE_WARNING) {
                } else {
                    Log::warn("External hook $hook->script exited with status code $result_code.", LogHook::EVENT_CODE_HOOK_NON_ZERO_EXIT_CODE_IN_WARN);
                }
            }
        }
    }
    abstract protected function getArgs($context);
}
interface HookContext {}
class FileHookContext implements HookContext
{
    public $share;
    public $path_on_share;
    public $from_path_on_share;
    public function __construct($share, $path_on_share, $from_path_on_share = NULL) {
        $this->share = $share;
        $this->path_on_share = $path_on_share;
        $this->from_path_on_share = $from_path_on_share;
    }
}
class FileHook extends Hook
{
    const EVENT_TYPE_CREATE = 'create';
    const EVENT_TYPE_EDIT   = 'edit';
    const EVENT_TYPE_RENAME = 'rename';
    const EVENT_TYPE_DELETE = 'delete';
    const EVENT_TYPE_MKDIR  = 'mkdir';
    const EVENT_TYPE_RMDIR  = 'rmdir';
    public static function trigger($event_type, $share, $path_on_share, $from_path_on_share = NULL) {
        Hook::_trigger($event_type, new FileHookContext($share, $path_on_share, $from_path_on_share));
    }
    protected function getArgs($context) {
        $args = array(
            escapeshellarg($this->event_type),
            escapeshellarg($context->share),
            escapeshellarg($context->path_on_share)
        );
        if (!empty($context->from_path_on_share)) {
            $args[] = escapeshellarg($context->from_path_on_share);
        }
        return $args;
    }
    public static function getEventTypes() {
        return array(
            static::EVENT_TYPE_CREATE,
            static::EVENT_TYPE_EDIT,
            static::EVENT_TYPE_RENAME,
            static::EVENT_TYPE_DELETE,
            static::EVENT_TYPE_MKDIR,
            static::EVENT_TYPE_RMDIR
        );
    }
}
class LogHookContext implements HookContext
{
    public $event_code;
    public $log;
    public function __construct($event_code, $log) {
        $this->event_code = $event_code;
        $this->log = $log;
    }
}
class LogHook extends Hook
{
    const EVENT_TYPE_WARNING  = 'warning';
    const EVENT_TYPE_ERROR    = 'error';
    const EVENT_TYPE_CRITICAL = 'critical';
    const EVENT_TYPE_IDLE     = 'idle';
    const EVENT_TYPE_NOT_IDLE = 'not_idle';
    const EVENT_TYPE_FSCK     = 'fsck';
    const EVENT_CODE_HOOK_NON_ZERO_EXIT_CODE_IN_WARN = 1; // Used to prevent infinite loop, when logging a WARNING from a LogHook!
    public static function trigger($event_type, $event_code, $log) {
        Hook::_trigger($event_type, new LogHookContext($event_code, $log));
    }
    protected function getArgs($context) {
        return array(
            escapeshellarg($this->event_type),
            escapeshellarg($context->event_code),
            escapeshellarg($context->log)
        );
    }
    public static function getEventTypes() {
        return array(
            static::EVENT_TYPE_WARNING,
            static::EVENT_TYPE_ERROR,
            static::EVENT_TYPE_CRITICAL,
            static::EVENT_TYPE_IDLE,
            static::EVENT_TYPE_NOT_IDLE,
            static::EVENT_TYPE_FSCK
        );
    }
}

define('ACTION_INITIALIZE', 'initialize');
define('ACTION_UNKNOWN', 'unknown');
define('ACTION_DAEMON', 'daemon');
define('ACTION_PAUSE', 'pause');
define('ACTION_RESUME', 'resume');
define('ACTION_FSCK', 'fsck');
define('ACTION_CANCEL_FSCK', 'cancel-fsck');
define('ACTION_BALANCE', 'balance');
define('ACTION_CANCEL_BALANCE', 'cancel-balance');
define('ACTION_STATS', 'stats');
define('ACTION_STATUS', 'status');
define('ACTION_LOGS', 'logs');
define('ACTION_EMPTY_TRASH', 'empty-trash');
define('ACTION_VIEW_QUEUE', 'view-queue');
define('ACTION_IOSTAT', 'iostat');
define('ACTION_GETUID', 'getuid');
define('ACTION_MD5_WORKER', 'md5-worker');
define('ACTION_FIX_SYMLINKS', 'fix-symlinks');
define('ACTION_REPLACE', 'replace');
define('ACTION_WAIT_FOR', 'wait-for');
define('ACTION_GONE', 'gone');
define('ACTION_GOING', 'going');
define('ACTION_THAW', 'thaw');
define('ACTION_DEBUG', 'debug');
define('ACTION_DELETE_METADATA', 'delete-metadata');
define('ACTION_REMOVE_SHARE', 'remove-share');
define('ACTION_CHECK_POOL', 'check_pool');
define('ACTION_SLEEP', 'sleep');
define('ACTION_READ_SAMBA_POOL', 'read_smb_spool');
define('ACTION_FSCK_FILE', 'fsck_file');
final class Log {
    const PERF     = 9;
    const TEST     = 8;
    const DEBUG    = 7;
    const INFO     = 6;
    const WARN     = 4;
    const ERROR    = 3;
    const CRITICAL = 2;
    const EVENT_CODE_ALL_DRIVES_FULL = 'all_drives_full';
    const EVENT_CODE_CONFIG_DEPRECATED_OPTION = 'config_deprecated_option';
    const EVENT_CODE_CONFIG_FILE_PARSING_FAILED = 'config_file_parsing_failed';
    const EVENT_CODE_CONFIG_HOOK_SCRIPT_NOT_EXECUTABLE = 'config_hook_script_not_executable';
    const EVENT_CODE_CONFIG_INCLUDE_INSECURE_PERMISSIONS = 'config_include_insecure_permissions';
    const EVENT_CODE_CONFIG_INVALID_VALUE = 'config_invalid_value';
    const EVENT_CODE_CONFIG_LZ_INSIDE_STORAGE_POOL = 'config_lz_inside_storage_pool';
    const EVENT_CODE_CONFIG_NO_STORAGE_POOL = 'config_no_storage_pool';
    const EVENT_CODE_CONFIG_SHARE_MISSING_FROM_SMB_CONF = 'config_share_missing_from_smb_conf';
    const EVENT_CODE_CONFIG_STORAGE_POOL_DRIVES_SAME_PARTITION = 'config_storage_pool_drives_same_partition';
    const EVENT_CODE_CONFIG_STORAGE_POOL_DRIVE_NOT_IN_DRIVE_SELECTION_ALGO = 'config_storage_pool_drive_not_in_drive_selection_algo';
    const EVENT_CODE_CONFIG_STORAGE_POOL_INSIDE_LZ = 'config_storage_pool_inside_lz';
    const EVENT_CODE_CONFIG_UNPARSEABLE_LINE = 'config_unparseable_line';
    const EVENT_CODE_DB_CONNECT_FAILED = 'db_connect_failed';
    const EVENT_CODE_DB_MIGRATION_FAILED = 'db_migration_failed';
    const EVENT_CODE_FILE_COPY_FAILED = 'file_copy_failed';
    const EVENT_CODE_FSCK_REPORT = 'fsck_report';
    const EVENT_CODE_FSCK_EMPTY_FILE_COPY_FOUND = 'fsck_empty_file_copy_found';
    const EVENT_CODE_FSCK_MD5_LOG_FAILURE = 'fsck_md5_log_failure';
    const EVENT_CODE_FSCK_MD5_MISMATCH = 'fsck_md5_mismatch';
    const EVENT_CODE_FSCK_METAFILE_ROOT_PATH_NOT_FOUND = 'fsck_metafile_root_path_not_found';
    const EVENT_CODE_FSCK_NO_FILE_COPIES = 'fsck_no_file_copies';
    const EVENT_CODE_FSCK_SIZE_MISMATCH_FILE_COPY_FOUND = 'fsck_size_mismatch_file_copy_found';
    const EVENT_CODE_FSCK_SYMLINK_FOUND_IN_STORAGE_POOL = 'fsck_symlink_found_in_storage_pool';
    const EVENT_CODE_FSCK_UNKNOWN_FOLDER = 'fsck_unknown_folder';
    const EVENT_CODE_FSCK_UNKNOWN_SHARE = 'fsck_unknown_share';
    const EVENT_CODE_HOOK_NON_ZERO_EXIT_CODE = 'hook_non_zero_exit_code';
    const EVENT_CODE_HOOK_NOT_EXECUTABLE = 'hook_not_executable';
    const EVENT_CODE_IDLE = "idle";
    const EVENT_CODE_IDLE_NOT = "busy";
    const EVENT_CODE_LIST_DIR_FAILED = 'list_dir_failed';
    const EVENT_CODE_MEMORY_LIMIT_REACHED = 'memory_limit_reached';
    const EVENT_CODE_METADATA_POINTS_TO_GONE_DRIVE = 'metadata_points_to_gone_drive';
    const EVENT_CODE_MKDIR_CHGRP_FAILED = 'mkdir_chgrp_failed';
    const EVENT_CODE_MKDIR_CHMOD_FAILED = 'mkdir_chmod_failed';
    const EVENT_CODE_MKDIR_CHOWN_FAILED = 'mkdir_chown_failed';
    const EVENT_CODE_MKDIR_FAILED = 'mkdir_failed';
    const EVENT_CODE_NO_METADATA_SAVED = 'no_metadata_saved';
    const EVENT_CODE_PHP_CRITICAL = 'php_critical';
    const EVENT_CODE_PHP_ERROR = 'php_error';
    const EVENT_CODE_PHP_WARNING = 'php_warning';
    const EVENT_CODE_RENAME_FILE_COPY_FAILED = 'rename_file_copy_failed';
    const EVENT_CODE_SAMBA_RESTART_FAILED = 'samba_restart_failed';
    const EVENT_CODE_SETTINGS_READ_ERROR = 'settings_read_error';
    const EVENT_CODE_SHARE_MISSING_FROM_GREYHOLE_CONF = 'share_missing_from_greyhole_conf';
    const EVENT_CODE_SPOOL_MOUNT_FAILED = 'spool_mount_failed';
    const EVENT_CODE_STORAGE_POOL_DRIVE_DF_FAILED = 'storage_pool_drive_df_failed';
    const EVENT_CODE_STORAGE_POOL_DRIVE_UUID_CHANGED = 'storage_pool_drive_uuid_changed';
    const EVENT_CODE_STORAGE_POOL_FOLDER_NOT_FOUND = 'storage_pool_folder_not_found';
    const EVENT_CODE_TASK_FOR_UNKNOWN_SHARE = 'task_for_unknown_share';
    const EVENT_CODE_TRASH_NOT_FOUND = 'trash_not_found';
    const EVENT_CODE_TRASH_SYMLINK_FAILED = 'trash_symlink_failed';
    const EVENT_CODE_UNEXPECTED_VAR = 'unexpected_var';
    const EVENT_CODE_VFS_MODULE_WRONG = 'vfs_module_wrong';
    const EVENT_CODE_ZFS_UNKNOWN_DEVICE = 'zfs_unknown_device';
    private static $log_level_names = array(
        9 => 'PERF',
        8 => 'TEST',
        7 => 'DEBUG',
        6 => 'INFO',
        4 => 'WARN',
        3 => 'ERROR',
        2 => 'CRITICAL',
    );
    private static $action = ACTION_INITIALIZE;
    private static $old_action;
    private static $level = Log::INFO; // Default, until we are able to read the config file
    public static function setLevel($level) {
        self::$level = $level;
    }
    public static function getLevel() {
        return self::$level;
    }
    public static function setAction($action) {
        self::$old_action = self::$action;
        self::$action = str_replace(':', '', $action);
    }
    public static function actionIs($action) {
        return self::$action == $action;
    }
    public static function restorePreviousAction() {
        self::$action = self::$old_action;
    }
    public static function debug($text, $event_code = NULL) {
        self::_log(self::DEBUG, $text, $event_code);
    }
    public static function info($text, $event_code = NULL) {
        self::_log(self::INFO, $text, $event_code);
    }
    public static function warn($text, $event_code) {
        self::_log(self::WARN, $text, $event_code);
    }
    public static function error($text, $event_code) {
        self::_log(self::ERROR, $text, $event_code);
    }
    public static function critical($text, $event_code) {
        self::_log(self::CRITICAL, $text, $event_code);
    }
    private static function _log($local_log_level, $text, $event_code) {
        if (self::$action == 'test-config') {
            $greyhole_log_file = NULL;
            $use_syslog = FALSE;
        } else {
            if ($local_log_level > self::$level) {
                return;
            }
            $greyhole_log_file = Config::get(CONFIG_GREYHOLE_LOG_FILE);
            $use_syslog = strtolower($greyhole_log_file) == 'syslog';
        }
        $date = date("M d H:i:s");
        if (self::$level >= self::PERF) {
            $utimestamp = microtime(true);
            $timestamp = floor($utimestamp);
            $date .= '.' . round(($utimestamp - $timestamp) * 1000000);
        }
        $log_level_string = $use_syslog ? $local_log_level : self::$log_level_names[$local_log_level];
        $log_text = sprintf("%s%s%s\n",
            "$date $log_level_string " . self::$action . ": ",
            $text,
            Config::get(CONFIG_LOG_MEMORY_USAGE) ? " [" . memory_get_usage() . "]" : ''
        );
        if ($use_syslog) {
            $worked = syslog($local_log_level, $log_text);
        } else if (!empty($greyhole_log_file)) {
            $worked = @error_log($log_text, 3, $greyhole_log_file);
            $greyhole_error_log_file = Config::get(CONFIG_GREYHOLE_ERROR_LOG_FILE);
            if ($local_log_level <= self::WARN && !empty($greyhole_error_log_file)) {
                $worked &= @error_log($log_text, 3, $greyhole_error_log_file);
            }
        } else {
            $worked = FALSE;
        }
        if (!$worked) {
            error_log(trim($log_text));
        }
        if ($local_log_level == self::WARN) {
            if ($event_code != LogHook::EVENT_CODE_HOOK_NON_ZERO_EXIT_CODE_IN_WARN) {
                LogHook::trigger(LogHook::EVENT_TYPE_WARNING, $event_code, $log_text);
            }
        } elseif ($local_log_level == self::ERROR) {
            LogHook::trigger(LogHook::EVENT_TYPE_ERROR, $event_code, $log_text);
        } elseif ($local_log_level == self::CRITICAL) {
            LogHook::trigger(LogHook::EVENT_TYPE_CRITICAL, $event_code, $log_text);
        }
        if ($local_log_level === self::CRITICAL) {
            exit(1);
        }
    }
}

final class MigrationHelper {
    public static function terminologyConversion() {
        self::convertFolders('.gh_graveyard', '.gh_metastore');
        self::convertFolders('.gh_graveyard_backup', '.gh_metastore_backup');
        self::convertFolders('.gh_attic', '.gh_trash');
        self::convertDatabase();
        self::convertStoragePoolDrivesTagFiles();
    }
	private static function convertFolders($old, $new) {
        foreach (Config::storagePoolDrives() as $sp_drive) {
			$old_term = clean_dir("$sp_drive/$old");
			$new_term = clean_dir("$sp_drive/$new");
			if (file_exists($old_term)) {
				Log::info("Moving $old_term to $new_term...");
				gh_rename($old_term, $new_term);
			}
		}
	}
	private static function convertDatabase() {
        Settings::rename('graveyard_backup_directory', 'metastore_backup_directory');
        $setting = Settings::get('metastore_backup_directory', FALSE, '%graveyard%');
        if ($setting) {
            $new_value = str_replace('/.gh_graveyard_backup', '/.gh_metastore_backup', $setting);
            Settings::set('metastore_backup_directory', $new_value);
        }
	}
    public static function convertStoragePoolDrivesTagFiles() {
        global $going_drive;
        $drives_definitions = Settings::get('sp_drives_definitions', TRUE);
        if (!$drives_definitions) {
            $drives_definitions = array();
        }
        foreach (Config::storagePoolDrives() as $sp_drive) {
            if (isset($going_drive) && $sp_drive == $going_drive) { continue; }
            $drive_uuid = SystemHelper::directory_uuid($sp_drive);
            if (!isset($drives_definitions[$sp_drive])) {
                if (is_dir($sp_drive)) {
                    $drives_definitions[$sp_drive] = $drive_uuid;
                }
            }
            if (!isset($drives_definitions[$sp_drive])) {
                continue;
            }
            if ($drives_definitions[$sp_drive] === FALSE) {
                unset($drives_definitions[$sp_drive]);
                continue;
            }
            if (file_exists("$sp_drive/.greyhole_uses_this") && $drive_uuid != 'remote') {
                unlink("$sp_drive/.greyhole_uses_this");
            }
            if ($drives_definitions[$sp_drive] != $drive_uuid) {
                Log::warn("Warning! It seems the partition UUID of $sp_drive changed. This probably means this mount is currently unmounted, or that you replaced this drive and didn't use 'greyhole --replace'. Because of that, Greyhole will NOT use this drive at this time.", Log::EVENT_CODE_STORAGE_POOL_DRIVE_UUID_CHANGED);
            }
        }
        foreach ($drives_definitions as $sp_drive => $uuid) {
            if (!array_contains(Config::storagePoolDrives(), $sp_drive)) {
                unset($drives_definitions[$sp_drive]);
            }
        }
        $devices = array();
        foreach ($drives_definitions as $sp_drive => $device_id) {
            $devices[$device_id][] = $sp_drive;
        }
        foreach ($devices as $device_id => $sp_drives) {
            if (count($sp_drives) > 1 && $device_id !== 0 && $device_id != 'remote') {
                if (Config::get(CONFIG_ALLOW_MULTIPLE_SP_PER_DRIVE)) {
                    Log::info("The following storage pool drives are on the same partition: " . implode(", ", $sp_drives) . ", but per greyhole.conf '" . CONFIG_ALLOW_MULTIPLE_SP_PER_DRIVE . "' options, you chose to ignore this normally critical error.");
                } else {
                    Log::critical("ERROR: The following storage pool drives are on the same partition: " . implode(", ", $sp_drives) . ". The Greyhole daemon will now stop.", Log::EVENT_CODE_CONFIG_STORAGE_POOL_DRIVES_SAME_PARTITION);
                }
            }
        }
        Settings::set('sp_drives_definitions', $drives_definitions);
        return $drives_definitions;
	}
}

class PoolDriveSelector {
    var $num_drives_per_draft;
    var $selection_algorithm;
    var $drives;
    var $is_forced;
    var $sorted_target_drives;
    var $last_resort_sorted_target_drives;
    function __construct($num_drives_per_draft, $selection_algorithm, $drives, $is_forced) {
        $this->num_drives_per_draft = $num_drives_per_draft;
        $this->selection_algorithm = $selection_algorithm;
        $this->drives = $drives;
        $this->is_forced = $is_forced;
    }
    public function isForced() {
        return $this->is_forced;
    }
    function init(&$sorted_target_drives, &$last_resort_sorted_target_drives) {
        if ($this->selection_algorithm == 'least_used_space') {
            $sorted_target_drives = $sorted_target_drives['used_space'];
            $last_resort_sorted_target_drives = $last_resort_sorted_target_drives['used_space'];
            asort($sorted_target_drives);
            asort($last_resort_sorted_target_drives);
        } else if ($this->selection_algorithm == 'most_available_space') {
            $sorted_target_drives = $sorted_target_drives['available_space'];
            $last_resort_sorted_target_drives = $last_resort_sorted_target_drives['available_space'];
            arsort($sorted_target_drives);
            arsort($last_resort_sorted_target_drives);
        } else {
            Log::critical("Unknown '" . CONFIG_DRIVE_SELECTION_ALGORITHM . "' found: " . $this->selection_algorithm, Log::EVENT_CODE_CONFIG_INVALID_VALUE);
        }
        $this->sorted_target_drives = array();
        foreach ($sorted_target_drives as $sp_drive => $space) {
            if (array_contains($this->drives, $sp_drive)) {
                $this->sorted_target_drives[$sp_drive] = $space;
            }
        }
        $this->last_resort_sorted_target_drives = array();
        foreach ($last_resort_sorted_target_drives as $sp_drive => $space) {
            if (array_contains($this->drives, $sp_drive)) {
                $this->last_resort_sorted_target_drives[$sp_drive] = $space;
            }
        }
    }
    function draft() {
        $drives = array();
        $drives_last_resort = array();
        while (count($drives)<$this->num_drives_per_draft) {
            $arr = kshift($this->sorted_target_drives);
            if ($arr === FALSE) {
                break;
            }
            list($sp_drive, $space) = $arr;
            if (!StoragePool::is_pool_drive($sp_drive)) { continue; }
            $drives[$sp_drive] = $space;
        }
        while (count($drives)+count($drives_last_resort)<$this->num_drives_per_draft) {
            $arr = kshift($this->last_resort_sorted_target_drives);
            if ($arr === FALSE) {
                break;
            }
            list($sp_drive, $space) = $arr;
            if (!StoragePool::is_pool_drive($sp_drive)) { continue; }
            $drives_last_resort[$sp_drive] = $space;
        }
        return array($drives, $drives_last_resort);
    }
    static function parse($config_string, $drive_selection_groups) {
        $ds = array();
        if ($config_string == 'least_used_space' || $config_string == 'most_available_space') {
            $ds[] = new PoolDriveSelector(count(Config::storagePoolDrives()), $config_string, Config::storagePoolDrives(), FALSE);
            return $ds;
        }
        if (!preg_match('/forced ?\((.+)\) ?(least_used_space|most_available_space)/i', $config_string, $regs)) {
            Log::critical("Can't understand the '" . CONFIG_DRIVE_SELECTION_ALGORITHM . "' value: $config_string", Log::EVENT_CODE_CONFIG_INVALID_VALUE);
        }
        $selection_algorithm = $regs[2];
        $groups = array_map('trim', explode(',', $regs[1]));
        foreach ($groups as $group) {
            $group = explode(' ', preg_replace('/^([0-9]+)x/', '\\1 ', $group));
            $num_drives = trim($group[0]);
            $group_name = trim($group[1]);
            if (!isset($drive_selection_groups[$group_name])) {
                continue;
            }
            if (stripos(trim($num_drives), 'all') === 0 || $num_drives > count($drive_selection_groups[$group_name])) {
                $num_drives = count($drive_selection_groups[$group_name]);
            }
            $ds[] = new PoolDriveSelector($num_drives, $selection_algorithm, $drive_selection_groups[$group_name], TRUE);
        }
        return $ds;
    }
    function update() {
        if (!$this->is_forced && ($this->selection_algorithm == 'least_used_space' || $this->selection_algorithm == 'most_available_space')) {
            $this->num_drives_per_draft = count(Config::storagePoolDrives());
            $this->drives = Config::storagePoolDrives();
        }
    }
}

final class Settings {
    public static function get($name, $unserialize=FALSE, $value=FALSE) {
        $query = "SELECT * FROM settings WHERE name LIKE :name";
        $params = array('name' => $name);
        if ($value !== FALSE) {
            $query .= " AND value LIKE :value";
            $params['value'] = $value;
        }
        $setting = DB::getFirst($query, $params);
        if ($setting === FALSE) {
            return FALSE;
        }
        return $unserialize ? unserialize($setting->value) : $setting->value;
    }
    public static function set($name, $value) {
        if (is_array($value)) {
            $value = serialize($value);
        }
        $query = "INSERT INTO settings SET name = :name, value = :value ON DUPLICATE KEY UPDATE value = VALUES(value)";
        DB::insert($query, array('name' => $name, 'value' => $value));
        return (object) array('name' => $name, 'value' => $value);
    }
    public static function rename($from, $to) {
        $query = "UPDATE settings SET name = :to WHERE name = :from";
        DB::execute($query, array('from' => $from, 'to' => $to));
    }
    public static function backup() {
        $settings = DB::getAll("SELECT * FROM settings");
        foreach (Config::storagePoolDrives() as $sp_drive) {
            if (StoragePool::is_pool_drive($sp_drive)) {
                $settings_backup_file = "$sp_drive/.gh_settings.bak";
                file_put_contents($settings_backup_file, serialize($settings));
            }
        }
    }
    public static function restore() {
        $latest_backup_time = 0;
        foreach (Config::storagePoolDrives() as $sp_drive) {
            $settings_backup_file = "$sp_drive/.gh_settings.bak";
            if (file_exists($settings_backup_file)) {
                $last_mod_date = filemtime($settings_backup_file);
                if ($last_mod_date > $latest_backup_time) {
                    $backup_file = $settings_backup_file;
                    $latest_backup_time = $last_mod_date;
                }
            }
        }
        if (isset($backup_file)) {
            Log::info("Restoring settings from last backup: $backup_file");
            $settings = unserialize(file_get_contents($backup_file));
            foreach ($settings as $setting) {
                Settings::set($setting->name, $setting->value);
            }
            return TRUE;
        }
        return FALSE;
    }
}

final class StoragePool {
    private static $greyhole_owned_drives = array();
    private static $gone_ok_drives = NULL;
    private static $fscked_gone_drives = NULL;
    public static function is_pool_drive($sp_drive) {
        global $going_drive;
        if (isset($going_drive) && $sp_drive == $going_drive) {
            return FALSE;
        }
        $is_greyhole_owned_drive = isset(self::$greyhole_owned_drives[$sp_drive]);
        if ($is_greyhole_owned_drive && self::$greyhole_owned_drives[$sp_drive] < time() - Config::get(CONFIG_DF_CACHE_TIME)) {
            unset(self::$greyhole_owned_drives[$sp_drive]);
            $is_greyhole_owned_drive = FALSE;
        }
        if (!$is_greyhole_owned_drive) {
            $drives_definitions = Settings::get('sp_drives_definitions', TRUE);
            if (!$drives_definitions) {
                $drives_definitions = MigrationHelper::convertStoragePoolDrivesTagFiles();
            }
            $drive_uuid = SystemHelper::directory_uuid($sp_drive);
            $is_greyhole_owned_drive = @$drives_definitions[$sp_drive] === $drive_uuid && $drive_uuid !== FALSE;
            if (!$is_greyhole_owned_drive) {
                $is_greyhole_owned_drive = file_exists("$sp_drive/.greyhole_uses_this");
                if ($is_greyhole_owned_drive && isset($drives_definitions[$sp_drive])) {
                    unset($drives_definitions[$sp_drive]);
                    Settings::set('sp_drives_definitions', $drives_definitions);
                }
            }
            if ($is_greyhole_owned_drive) {
                self::$greyhole_owned_drives[$sp_drive] = time();
            }
        }
        return $is_greyhole_owned_drive;
    }
    public static function check_drives($skip_fsck=FALSE) {
        Log::setAction(ACTION_CHECK_POOL);
        $needs_fsck = FALSE;
        $drives_definitions = Settings::get('sp_drives_definitions', TRUE);
        $returned_drives = array();
        $missing_drives = array();
        $i = 0; $j = 0;
        foreach (Config::storagePoolDrives() as $sp_drive) {
            if (!self::is_pool_drive($sp_drive) && !self::gone_fscked($sp_drive, $i++ == 0) && !file_exists("$sp_drive/.greyhole_used_this") && !empty($drives_definitions[$sp_drive])) {
                if($needs_fsck !== 2){
                    $needs_fsck = 1;
                }
                self::mark_gone_drive_fscked($sp_drive);
                $missing_drives[] = $sp_drive;
                Log::warn("Warning! It seems the partition UUID of $sp_drive changed. This probably means this mount is currently unmounted, or that you replaced this drive and didn't use 'greyhole --replace'. Because of that, Greyhole will NOT use this drive at this time.", Log::EVENT_CODE_STORAGE_POOL_DRIVE_UUID_CHANGED);
                Log::debug("Email sent for gone drive: $sp_drive");
                self::$gone_ok_drives[$sp_drive] = TRUE; // The upcoming fsck should not recreate missing copies just yet
            } else if ((self::gone_ok($sp_drive, $j++ == 0) || self::gone_fscked($sp_drive, $i++ == 0)) && self::is_pool_drive($sp_drive) && !empty($drives_definitions[$sp_drive])) {
                $needs_fsck = 2;
                $returned_drives[] = $sp_drive;
                Log::debug("Email sent for revived drive: $sp_drive");
                self::mark_gone_ok($sp_drive, 'remove');
                self::mark_gone_drive_fscked($sp_drive, 'remove');
                $i = 0; $j = 0;
            }
        }
        if (count($returned_drives) > 0) {
            $body = "This is an automated email from Greyhole.\n\nOne (or more) of your storage pool drives came back:\n";
            foreach ($returned_drives as $sp_drive) {
                $body .= "$sp_drive was missing; it's now available again.\n";
            }
            if (!$skip_fsck) {
                $body .= "\nA fsck will now start, to fix the symlinks found in your shares, when possible.\nYou'll receive a report email once that fsck run completes.\n";
            }
            $drive_string = join(", ", $returned_drives);
            $subject = "Storage pool drive now online on " . exec ('hostname') . ": ";
            $subject = $subject . $drive_string;
            if (strlen($subject) > 255) {
                $subject = substr($subject, 0, 255);
            }
            mail(Config::get(CONFIG_EMAIL_TO), $subject, $body);
        }
        if (count($missing_drives) > 0) {
            $body = "This is an automated email from Greyhole.\n\nOne (or more) of your storage pool drives has disappeared:\n";
            foreach ($missing_drives as $sp_drive) {
                if (!is_dir($sp_drive)) {
                    $body .= "$sp_drive: directory doesn't exists\n";
                } else {
                    $current_uuid = SystemHelper::directory_uuid($sp_drive);
                    if (empty($current_uuid)) {
                        $current_uuid = 'N/A';
                    }
                    $body .= "$sp_drive: expected partition UUID: " . $drives_definitions[$sp_drive] . "; current partition UUID: $current_uuid\n";
                }
            }
            $sp_drive = $missing_drives[0];
            $body .= "\nThis either means this mount is currently unmounted, or you forgot to use 'greyhole --replace' when you changed this drive.\n\n";
            $body .= "Here are your options:\n\n";
            $body .= "- If you forgot to use 'greyhole --replace', you should do so now. Until you do, this drive will not be part of your storage pool.\n\n";
            $body .= "- If the drive is gone, you should either re-mount it manually (if possible), or remove it from your storage pool. To do so, use the following command:\n  greyhole --gone=" . escapeshellarg($sp_drive) . "\n  Note that the above command is REQUIRED for Greyhole to re-create missing file copies before the next fsck runs. Until either happens, missing file copies WILL NOT be re-created on other drives.\n\n";
            $body .= "- If you know this drive will come back soon, and do NOT want Greyhole to re-create missing file copies for this drive until it reappears, you should execute this command:\n  greyhole --wait-for=" . escapeshellarg($sp_drive) . "\n\n";
            if (!$skip_fsck) {
                $body .= "A fsck will now start, to fix the symlinks found in your shares, when possible.\nYou'll receive a report email once that fsck run completes.\n";
            }
            $subject = "Missing storage pool drives on " . exec('hostname') . ": ";
            $drive_string = join(",",$missing_drives);
            $subject = $subject . $drive_string;
            if (strlen($subject) > 255) {
                $subject = substr($subject, 0, 255);
            }
            mail(Config::get(CONFIG_EMAIL_TO), $subject, $body);
        }
        if ($needs_fsck !== FALSE) {
            set_metastore_backup();
            get_metastores(FALSE); // FALSE => Resets the metastores cache
            clearstatcache();
            if (!$skip_fsck) {
                initialize_fsck_report('All shares');
                if ($needs_fsck === 2) {
                    foreach ($returned_drives as $drive) {
                        $metastores = get_metastores_from_storage_volume($drive);
                        Log::info("Starting fsck for metadata store on $drive which came back online.");
                        foreach ($metastores as $metastore) {
                            foreach (SharesConfig::getShares() as $share_name => $share_options) {
                                gh_fsck_metastore($metastore,"/$share_name", $share_name);
                            }
                        }
                        Log::info("fsck for returning drive $drive's metadata store completed.");
                    }
                    Log::info("Starting fsck for all shares - caused by missing drive that came back online.");
                } else {
                    Log::info("Starting fsck for all shares - caused by missing drive. Will just recreate symlinks to existing copies when possible; won't create new copies just yet.");
                    fix_all_symlinks();
                }
                schedule_fsck_all_shares(array('email'));
                Log::info("  fsck for all shares scheduled.");
            }
            self::reload_gone_ok_drives();
        }
    }
    public static function gone_ok($sp_drive, $force_reload=FALSE) {
        if ($force_reload || self::$gone_ok_drives === NULL) {
            self::reload_gone_ok_drives();
        }
        if (isset(self::$gone_ok_drives[$sp_drive])) {
            return TRUE;
        }
        return FALSE;
    }
    public static function reload_gone_ok_drives() {
        self::$gone_ok_drives = Settings::get('Gone-OK-Drives', TRUE);
        if (!self::$gone_ok_drives) {
            self::$gone_ok_drives = array();
            Settings::set('Gone-OK-Drives', self::$gone_ok_drives);
        }
    }
    public static function get_gone_ok_drives() {
        if (self::$gone_ok_drives === NULL) {
            self::reload_gone_ok_drives();
        }
        return self::$gone_ok_drives;
    }
    public static function mark_gone_ok($sp_drive, $action='add') {
        if (!array_contains(Config::storagePoolDrives(), $sp_drive)) {
            $sp_drive = '/' . trim($sp_drive, '/');
        }
        if (!array_contains(Config::storagePoolDrives(), $sp_drive)) {
            return FALSE;
        }
        self::reload_gone_ok_drives();
        if ($action == 'add') {
            self::$gone_ok_drives[$sp_drive] = TRUE;
        } else {
            unset(self::$gone_ok_drives[$sp_drive]);
        }
        Settings::set('Gone-OK-Drives', self::$gone_ok_drives);
        return TRUE;
    }
    public static function gone_fscked($sp_drive, $force_reload=FALSE) {
        if ($force_reload || self::$fscked_gone_drives == NULL) {
            self::reload_fsck_gone_drives();
        }
        if (isset(self::$fscked_gone_drives[$sp_drive])) {
            return TRUE;
        }
        return FALSE;
    }
    public static function reload_fsck_gone_drives() {
        self::$fscked_gone_drives = Settings::get('Gone-FSCKed-Drives', TRUE);
        if (!self::$fscked_gone_drives) {
            self::$fscked_gone_drives = array();
            Settings::set('Gone-FSCKed-Drives', self::$fscked_gone_drives);
        }
    }
    public static function mark_gone_drive_fscked($sp_drive, $action='add') {
        self::reload_fsck_gone_drives();
        if ($action == 'add') {
            self::$fscked_gone_drives[$sp_drive] = TRUE;
        } else {
            unset(self::$fscked_gone_drives[$sp_drive]);
        }
        Settings::set('Gone-FSCKed-Drives', self::$fscked_gone_drives);
    }
}

final class SystemHelper {
    public static function is_amahi() {
        return file_exists('/usr/bin/hda-ctl');
    }
    public static function directory_uuid($dir) {
        $dev = exec('df ' . escapeshellarg($dir) . ' 2> /dev/null | grep \'/dev\' | awk \'{print $1}\'');
        if (!is_dir($dir)) {
            return FALSE;
        }
        if (empty($dev) || strpos($dev, '/dev/') !== 0) {
            if (file_exists('/sbin/zpool')) {
                $dataset = exec('df ' . escapeshellarg($dir) . ' 2> /dev/null | awk \'{print $1}\'');
                if (strpos($dataset, '/') !== FALSE) {
                    $is_zfs = exec('mount | grep ' . escapeshellarg("$dataset .*zfs") . ' 2> /dev/null | wc -l');
                    if ($is_zfs == 1) {
                        $p = explode('/', $dataset);
                        $pool = $p[0];
                        $dev_name = exec('/sbin/zpool list -v ' . escapeshellarg($pool) . ' 2> /dev/null | awk \'{print $1}\' | tail -n 1');
                        if (!empty($dev_name)) {
                            $dev = exec("ls -l /dev/disk/*/$dev_name | awk '{print \$(NF-2)}'");
                            if (empty($dev) && file_exists("/dev/$dev_name")) {
                                $dev = '/dev/$dev_name';
                                Log::info("Found a ZFS pool ($pool) that uses a device name in /dev/ ($dev). That is a bad idea, since those can easily change, which would prevent this pool from mounting automatically. You should use any of the /dev/disk/*/ links instead. For example, you could do: zpool export $pool && zpool import -d /dev/disk/by-id/ $pool. More details at http://zfsonlinux.org/faq.html#WhatDevNamesShouldIUseWhenCreatingMyPool");
                            }
                        }
                        if (empty($dev)) {
                            Log::warn("Warning! Couldn't find the device used by your ZFS pool name $pool. That pool will never be used.", Log::EVENT_CODE_ZFS_UNKNOWN_DEVICE);
                            return FALSE;
                        }
                    }
                }
            }
            if (empty($dev)) {
                return 'remote';
            }
        }
        $uuid = trim(exec('/sbin/blkid '.$dev.' | awk -F\'UUID="\' \'{print $2}\' | awk -F\'"\' \'{print $1}\''));
        if (empty($uuid)) {
            return 'remote';
        }
        return $uuid;
    }
}
$use_alt_symlinks_creation = FALSE;
function gh_symlink($target, $link) {
    global $use_alt_symlinks_creation;
    $success = !$use_alt_symlinks_creation && symlink($target, $link);
    if (!$success) {
        exec("ln -s " . escapeshellarg($target) . " " . escapeshellarg($link));
        $success = gh_is_file($link);
        if ($success) {
            if (!$use_alt_symlinks_creation) {
                Log::info("Will use exec() instead of symlink() to create all symlinks.");
            }
            $use_alt_symlinks_creation = TRUE;
        }
    }
    return $success;
}
$arch = exec('uname -m');
if ($arch != 'x86_64') {
    function gh_filesize($filename) {
        $result = exec("stat -c %s ".escapeshellarg($filename)." 2>/dev/null");
        if (empty($result)) {
            return FALSE;
        }
        return (float) $result;
    }
    function gh_fileowner($filename) {
        $result = exec("stat -c %u ".escapeshellarg($filename)." 2>/dev/null");
        if (empty($result)) {
            return FALSE;
        }
        return (int) $result;
    }
    function gh_filegroup($filename) {
        $result = exec("stat -c %g ".escapeshellarg($filename)." 2>/dev/null");
        if (empty($result)) {
            return FALSE;
        }
        return (int) $result;
    }
    function gh_fileperms($filename) {
        $result = exec("stat -c %a ".escapeshellarg($filename)." 2>/dev/null");
        if (empty($result)) {
            return FALSE;
        }
        return "0" . $result;
    }
    function gh_is_file($filename) {
        exec('[ -f '.escapeshellarg($filename).' ]', $tmp, $result);
        return $result === 0;
    }
    function gh_fileinode($filename) {
        $result = exec("stat -c '%d_%i' ".escapeshellarg($filename)." 2>/dev/null");
        if (empty($result)) {
            return FALSE;
        }
        return (string) $result;
    }
    function gh_file_deviceid($filename) {
        $result = exec("stat -c '%d' ".escapeshellarg($filename)." 2>/dev/null");
        if (empty($result)) {
            return FALSE;
        }
        return (string) $result;
    }
    function gh_rename($filename, $target_filename) {
        exec("mv ".escapeshellarg($filename)." ".escapeshellarg($target_filename)." 2>/dev/null", $output, $result);
        return $result === 0;
    }
} else {
    function gh_filesize(&$filename) {
        $size = @filesize($filename);
        if ($size === FALSE) {
            $size = @filesize(normalize_utf8_characters($filename));
            if ($size !== FALSE) {
                $filename = normalize_utf8_characters($filename);
            }
        }
        return $size;
    }
    function gh_fileowner($filename) {
        return fileowner($filename);
    }
    function gh_filegroup($filename) {
        return filegroup($filename);
    }
    function gh_fileperms($filename) {
        return mb_substr(decoct(fileperms($filename)), -4);
    }
    function gh_is_file($filename) {
        return is_file($filename);
    }
    function gh_fileinode($filename) {
        $stat = @stat($filename);
        if ($stat === FALSE) {
            return FALSE;
        }
        return $stat['dev'] . '_' . $stat['ino'];
    }
    function gh_file_deviceid($filename) {
        $stat = @stat($filename);
        if ($stat === FALSE) {
            return FALSE;
        }
        return $stat['dev'];
    }
    function gh_rename($filename, $target_filename) {
        return @rename($filename, $target_filename);
    }
}

$constarray = get_defined_constants(true);
foreach($constarray['user'] as $key => $val) {
    eval(sprintf('$_CONSTANTS[\'%s\'] = ' . (is_int($val) || is_float($val) ? '%s' : "'%s'") . ';', addslashes($key), addslashes($val)));
}
define('FSCK_TYPE_SHARE', 1);
define('FSCK_TYPE_STORAGE_POOL_DRIVE', 2);
define('FSCK_TYPE_METASTORE', 3);
set_error_handler("gh_error_handler");
register_shutdown_function("gh_shutdown");
umask(0);
setlocale(LC_COLLATE, "en_US.UTF-8");
setlocale(LC_CTYPE, "en_US.UTF-8");
if (!defined('PHP_VERSION_ID')) {
    $version = explode('.', PHP_VERSION);
    define('PHP_VERSION_ID', ($version[0] * 10000 + $version[1] * 100 + $version[2]));
}
$sleep_before_task = array();
function clean_dir($dir) {
    if (empty($dir)) {
        return $dir;
    }
    if ($dir[0] == '.' && $dir[1] == '/') {
        $dir = mb_substr($dir, 2);
    }
    while (string_contains($dir, '//')) {
        $dir = str_replace("//", "/", $dir);
    }
    $l = strlen($dir);
    if ($l >= 2 && $dir[$l-2] == '/' && $dir[$l-1] == '.') {
        $dir = mb_substr($dir, 0, $l-2);
    }
    $dir = str_replace("/./", "/", $dir);
    return $dir;
}
function explode_full_path($full_path) {
    return array(dirname($full_path), basename($full_path));
}
function gh_shutdown() {
    if ($err = error_get_last()) {
        Log::error("PHP Fatal Error: " . $err['message'] . "; BT: " . basename($err['file']) . '[L' . $err['line'] . '] ', Log::EVENT_CODE_PHP_CRITICAL);
    }
}
function gh_error_handler($errno, $errstr, $errfile, $errline, $errcontext) {
    if(!($errno & error_reporting())) {
        return TRUE;
    }
    switch ($errno) {
    case E_ERROR:
    case E_PARSE:
    case E_CORE_ERROR:
    case E_COMPILE_ERROR:
        Log::critical("PHP Error [$errno]: $errstr in $errfile on line $errline", Log::EVENT_CODE_PHP_CRITICAL);
        break;
    case E_WARNING:
    case E_COMPILE_WARNING:
    case E_CORE_WARNING:
    case E_NOTICE:
        $greyhole_log_file = Config::get(CONFIG_GREYHOLE_LOG_FILE);
        if ($errstr == "fopen($greyhole_log_file): failed to open stream: Permission denied") {
            return TRUE;
        }
        Log::warn("PHP Warning [$errno]: $errstr in $errfile on line $errline; BT: " . get_debug_bt(), Log::EVENT_CODE_PHP_WARNING);
        break;
    default:
        Log::error("PHP Unknown Error [$errno]: $errstr in $errfile on line $errline", Log::EVENT_CODE_PHP_ERROR);
        break;
    }
    return TRUE;
}
function get_debug_bt() {
    $bt = '';
    foreach (debug_backtrace() as $d) {
        if ($d['function'] == 'gh_error_handler' || $d['function'] == 'get_debug_bt') { continue; }
        if ($bt != '') {
            $bt = " => $bt";
        }
        $prefix = '';
        if (isset($d['file'])) {
            $prefix = basename($d['file']) . '[L' . $d['line'] . '] ';
        }
        if (!isset($d['args'])) {
            $d['args'] = array();
        }
        foreach ($d['args'] as $k => $v) {
            if (is_object($v)) {
                $d['args'][$k] = 'stdClass';
            }
            if (is_array($v)) {
                $d['args'][$k] = str_replace("\n", "", var_export($v, TRUE));
            }
        }
        $bt = $prefix . $d['function'] .'(' . implode(',', @$d['args']) . ')' . $bt;
    }
    return $bt;
}
function bytes_to_human($bytes, $html=TRUE) {
    $units = 'B';
    if (abs($bytes) > 1024) {
        $bytes /= 1024;
        $units = 'KB';
    }
    if (abs($bytes) > 1024) {
        $bytes /= 1024;
        $units = 'MB';
    }
    if (abs($bytes) > 1024) {
        $bytes /= 1024;
        $units = 'GB';
    }
    if (abs($bytes) > 1024) {
        $bytes /= 1024;
        $units = 'TB';
    }
    $decimals = (abs($bytes) > 100 ? 0 : (abs($bytes) > 10 ? 1 : 2));
    if ($html) {
        return number_format($bytes, $decimals) . " <span class=\"i18n-$units\">$units</span>";
    } else {
        return number_format($bytes, $decimals) . $units;
    }
}
function duration_to_human($seconds) {
    $displayable_duration = '';
    if ($seconds > 60*60) {
        $hours = floor($seconds / (60*60));
        $displayable_duration .= $hours . 'h ';
        $seconds -= $hours * (60*60);
    }
    if ($seconds > 60) {
        $minutes = floor($seconds / 60);
        $displayable_duration .= $minutes . 'm ';
        $seconds -= $minutes * 60;
    }
    $displayable_duration .= $seconds . 's';
    return $displayable_duration;
}
function get_share_landing_zone($share) {
    $lz = SharesConfig::get($share, CONFIG_LANDING_ZONE);
    if ($lz !== FALSE) {
        return $lz;
    } else if (array_contains(ConfigHelper::$trash_share_names, $share)) {
        return SharesConfig::get(CONFIG_TRASH_SHARE, CONFIG_LANDING_ZONE);
    } else {
        Log::warn("  Found a share ($share) with no path in " . ConfigHelper::$smb_config_file . ", or missing it's num_copies[$share] config in " . ConfigHelper::$config_file . ". Skipping.", Log::EVENT_CODE_SHARE_MISSING_FROM_GREYHOLE_CONF);
        return FALSE;
    }
}
function memory_check() {
    $usage = memory_get_usage();
    $used = $usage / Config::get(CONFIG_MEMORY_LIMIT);
    $used = $used * 100;
    if ($used > 95) {
        Log::critical("$used% memory usage, exiting. Please increase '" . CONFIG_MEMORY_LIMIT . "' in /etc/greyhole.conf", Log::EVENT_CODE_MEMORY_LIMIT_REACHED);
    }
}
class metafile_iterator implements Iterator {
    private $path;
    private $share;
    private $load_nok_metafiles;
    private $quiet;
    private $check_symlink;
    private $metafiles;
    private $metastores;
    private $dir_handle;
    private $directory_stack;
    public function __construct($share, $path, $load_nok_metafiles=FALSE, $quiet=FALSE, $check_symlink=TRUE) {
        $this->quiet = $quiet;
        $this->share = $share;
        $this->path = $path;
        $this->check_symlink = $check_symlink;
        $this->load_nok_metafiles = $load_nok_metafiles;
    }
    public function rewind() {
        $this->metastores = get_metastores();
        $this->directory_stack = array($this->path);
        $this->dir_handle = NULL;
        $this->metafiles = array();
        $this->next();
    }
    public function current() {
        return $this->metafiles;
    }
    public function key() {
        return count($this->metafiles);
    }
    public function next() {
        $this->metafiles = array();
        while (count($this->directory_stack) > 0 && $this->directory_stack !== NULL) {
            $dir = array_pop($this->directory_stack);
            if (!$this->quiet) {
                Log::debug("Loading metadata files for (dir) " . clean_dir($this->share . (!empty($dir) ? "/" . $dir : "")) . " ...");
            }
            for ($i = 0; $i < count($this->metastores); $i++) {
                $metastore = $this->metastores[$i];
                $base = "$metastore/" . $this->share . "/";
                if (!file_exists($base . $dir)) {
                    continue;
                }    
                if ($this->dir_handle = opendir($base . $dir)) {
                    while (false !== ($file = readdir($this->dir_handle))) {
                        memory_check();
                        if ($file=='.' || $file=='..') {
                            continue;
                        }
                        if (!empty($dir)) {
                            $full_filename = $dir . '/' . $file;
                        } else {
                            $full_filename = $file;
                        }
                        if (is_dir($base . $full_filename)) {
                            $this->directory_stack[] = $full_filename;
                        } else {
                            $full_filename = str_replace("$this->path/",'',$full_filename);
                            if (isset($this->metafiles[$full_filename])) {
                                continue;
                            }                        
                            $this->metafiles[$full_filename] = get_metafiles_for_file($this->share, $dir, $file, $this->load_nok_metafiles, $this->quiet, $this->check_symlink);
                        }
                    }
                    closedir($this->dir_handle);
                    $this->directory_stack = array_unique($this->directory_stack);
                }
            }
            if (count($this->metafiles) > 0) {
                break;
            }
        }
        if (!$this->quiet) {
            Log::debug('Found ' . count($this->metafiles) . ' metadata files.');
        }
        return $this->metafiles;
    }
    public function valid() {
        return count($this->metafiles) > 0;
    }
}
class FSCKLogFile {
    const PATH = '/usr/share/greyhole';
    private $path;
    private $filename;
    private $lastEmailSentTime = 0;
    public function __construct($filename, $path=self::PATH) {
        $this->filename = $filename;
        $this->path = $path;
    }
    public function emailAsRequired() {
        $logfile = "$this->path/$this->filename";
        if (!file_exists($logfile)) { return; }
        $last_mod_date = filemtime($logfile);
        if ($last_mod_date > $this->getLastEmailSentTime()) {
            if ($this->shouldSendViaEmail()) {
                $email_to = Config::get(CONFIG_EMAIL_TO);
                Log::info("Sending $logfile by email to $email_to");
                mail($email_to, $this->getSubject(), $this->getBody());
            }
            LogHook::trigger(LogHook::EVENT_TYPE_FSCK, Log::EVENT_CODE_FSCK_REPORT, $this->getSubject() . "\n" . $this->getBody());
            $this->lastEmailSentTime = $last_mod_date;
            Settings::set("last_email_$this->filename", $this->lastEmailSentTime);
        }
    }
    private function shouldSendViaEmail() {
        $this->getBody();
        global $fsck_report;
        return (@$fsck_report['send_via_email'] === TRUE);
    }
    private function getBody() {
        $logfile = "$this->path/$this->filename";
        if ($this->filename == 'fsck_checksums.log') {
            return file_get_contents($logfile) . "\nNote: You should manually delete the $logfile file once you're done with it.";
        } else if ($this->filename == 'fsck_files.log') {
            global $fsck_report;
            $fsck_report = unserialize(file_get_contents($logfile));
            unlink($logfile);
            return get_fsck_report() . "\nNote: This report is a complement to the last report you've received. It details possible errors with files for which the fsck was postponed.";
        } else {
            return '[empty]';
        }
    }
    private function getSubject() {
        if ($this->filename == 'fsck_checksums.log') {
            return 'Mismatched checksums in Greyhole file copies';
        } else if ($this->filename == 'fsck_files.log') {
            return 'fsck_files of Greyhole shares on ' . exec('hostname');
        } else {
            return 'Unknown FSCK report';
        }
    }
    private function getLastEmailSentTime() {
        if ($this->lastEmailSentTime == 0) {
            $setting = Settings::get("last_email_$this->filename");
            if ($setting) {
                $this->lastEmailSentTime = (int) $setting;
            }
        }
        return $this->lastEmailSentTime;
    }
    public static function loadFSCKReport($what) {
        $logfile = self::PATH . '/fsck_files.log';
        if (file_exists($logfile)) {
            global $fsck_report;
            $fsck_report = unserialize(file_get_contents($logfile));
            if ($fsck_report !== FALSE) {
                return;
            }
            rename($logfile, "$logfile.broken");
        }
        initialize_fsck_report($what);
    }
    public static function saveFSCKReport($send_via_email) {
        global $fsck_report;
        $fsck_report['send_via_email'] = $send_via_email;
        $logfile = self::PATH . '/fsck_files.log';
        file_put_contents($logfile, serialize($fsck_report));
    }
}
function fix_all_symlinks() {
    foreach (SharesConfig::getShares() as $share_name => $share_options) {
        fix_symlinks_on_share($share_name);
    }
}
function fix_symlinks_on_share($share_name) {
    $share_options = SharesConfig::getConfigForShare($share_name);
    echo "Looking for broken symbolic links in the share '$share_name'...";
    chdir($share_options[CONFIG_LANDING_ZONE]);
    exec("find -L . -type l", $result);
    foreach ($result as $file_to_relink) {
        if (is_link($file_to_relink)) {
            $file_to_relink = substr($file_to_relink, 2);
            foreach (Config::storagePoolDrives() as $sp_drive) {
                if (!StoragePool::is_pool_drive($sp_drive)) { continue; }
                $new_link_target = clean_dir("$sp_drive/$share_name/$file_to_relink");
                if (gh_is_file($new_link_target)) {
                    unlink($file_to_relink);
                    gh_symlink($new_link_target, $file_to_relink);
                    break;
                }
            }
        }
    }
    echo " Done.\n";
}
function schedule_fsck_all_shares($fsck_options=array()) {
    foreach (SharesConfig::getShares() as $share_name => $share_options) {
        $query = "INSERT INTO tasks SET action = 'fsck', share = :full_path, additional_info = :fsck_options, complete = 'yes'";
        $params = array(
            'full_path' => $share_options[CONFIG_LANDING_ZONE],
            'fsck_options' => empty($fsck_options) ? NULL : implode('|', $fsck_options)
        );
        DB::insert($query, $params);
    }
}
function kshift(&$arr) {
    if (count($arr) == 0) {
        return FALSE;
    }
    foreach ($arr as $k => $v) {
        unset($arr[$k]);
        break;
    }
    return array($k, $v);
}
function array_contains($haystack, $needle) {
    return array_search($needle, $haystack) !== FALSE;
}
function string_contains($haystack, $needle) {
    return mb_strpos($haystack, $needle) !== FALSE;
}
function string_starts_with($haystack, $needle) {
    return mb_strpos($haystack, $needle) === 0;
}
function json_pretty_print($json) {
    if (!is_string($json)) {
        $json = json_encode($json);
    }
    $result = '';
    $level = 0;
    $in_quotes = FALSE;
    $in_escape = FALSE;
    $ends_line_level = NULL;
    $json_length = strlen( $json );
    for ($i = 0; $i < $json_length; $i++) {
        $char = $json[$i];
        $new_line_level = NULL;
        $post = "";
        if ($ends_line_level !== NULL) {
            $new_line_level = $ends_line_level;
            $ends_line_level = NULL;
        }
        if ($in_escape) {
            $in_escape = FALSE;
        } else if ($char === '"') {
            $in_quotes = !$in_quotes;
        } else if (!$in_quotes) {
            switch($char) {
                case '}':
                case ']':
                    $level--;
                    $ends_line_level = NULL;
                    $new_line_level = $level;
                    break;
                case '{':
                case '[':
                    $level++;
                case ',':
                    $ends_line_level = $level;
                    break;
                case ':':
                    $post = " ";
                    break;
                case " ":
                case "\t":
                case "\n":
                case "\r":
                    $char = "";
                    $ends_line_level = $new_line_level;
                    $new_line_level = NULL;
                    break;
            }
        } else if ($char === '\\') {
            $in_escape = true;
        }
        if ($new_line_level !== NULL) {
            $result .= "\n" . str_repeat("  ", $new_line_level);
        }
        $result .= $char . $post;
    }
    return $result;
}

ConfigHelper::parse();
DB::connect();
header('Content-Type: text/html; charset=utf8');
setlocale(LC_CTYPE, "en_US.UTF-8");
if (isset($_REQUEST['level'])) {
    $level_min = (int) $_REQUEST['level'];
} else {
    $level_min = 1;
}
$level_max = $level_min+1;
if (isset($_REQUEST['path'])) {
    $path = $_REQUEST['path'];
} else {
    $path = '/';
}
$p = explode('/', trim($path, '/'));
$share = array_shift($p);
$full_path = implode('/', $p);
if ($level_min > 1) {
	$query = "SELECT size FROM du_stats WHERE depth = :depth AND share = :share AND full_path = :full_path";
    $params = array(
		'depth' => $level_min - 1,
		'share' => $share,
		'full_path' => $full_path
	);
    $total_bytes = (float) DB::getFirstValue($query, $params) or die("SQL error: " . DB::error());
}
if ($level_min > 1) {
	$query = "SELECT size, depth, CONCAT('/', share, '/', full_path) AS file_path FROM du_stats WHERE depth = :depth AND share = :share AND full_path LIKE :full_path";
    $params = array(
		'depth' => $level_min,
		'share' => $share,
		'full_path' => empty($full_path) ? '%' : "$full_path/%"
	);
} else {
	$query = "SELECT size, depth, CONCAT('/', share, '/', full_path) AS file_path FROM du_stats WHERE depth = :depth";
    $params = array(
		'depth' => $level_min
	);
}
$rows = DB::getAll($query, $params) or die("SQL error: " . DB::error());
$total_bytes_subfolders = 0;
$results_rows = array();
foreach ($rows as $row) {
	$results_rows[] = $row;
    if ($row->depth == $level_min) {
        $total_bytes_subfolders += (float) $row->size;
    }
}
if ($level_min == 1) {
    $total_bytes = $total_bytes_subfolders;
}
$total_bytes_files = $total_bytes - $total_bytes_subfolders;
function escape($string) {
    return str_replace("'", "\\'", $string);
}
?>
<html>
  <head>
    <script type="text/javascript" src="https://www.google.com/jsapi"></script>
    <script type="text/javascript">
      google.load("visualization", "1", {packages:["treemap"]});
      google.setOnLoadCallback(drawChart);
      function drawChart() {
        var data = google.visualization.arrayToDataTable([
          ['Path', 'Parent', 'Size (bytes)', 'Color'],
          ['<?php echo escape($path) ?>', null, <?php echo $total_bytes ?>, <?php echo $total_bytes ?>],
          ['Files', '<?php echo escape($path) ?>', <?php echo $total_bytes_files ?>, <?php echo $total_bytes_files ?>],
          <?php
          foreach ($results_rows as $row) {
              $bytes = (float) $row->size;
              $parent = dirname($row->file_path);
              if ($row->file_path[strlen($row->file_path)-1] == '/') {
                  $row->file_path = substr($row->file_path, 0, strlen($row->file_path)-1);
              }
              echo "['" . escape($row->file_path) . "','".escape($parent)."',$bytes,$bytes],\n";
          }
          ?>
        ]);
        var text_data = [
            {text: '<?php echo escape($path) ?> (<?php echo escape(bytes_to_human($total_bytes, FALSE)) ?>)', path: '<?php echo escape($path) ?>'},
            {text: 'Files (<?php echo escape(bytes_to_human($total_bytes_files, FALSE)) ?>)', path: 'Files'},
        <?php
        foreach ($results_rows as $row) {
            $bytes = (float) $row->size;
            echo "{text:'" . escape($row->file_path) . " (".escape(bytes_to_human($bytes, FALSE)).")', path: '" . escape($row->file_path) . "'},\n";
        }
        ?>
        ];
        var tree = new google.visualization.TreeMap(document.getElementById('chart_div'));
        tree.draw(data, {
          headerHeight: 15,
          fontColor: 'black'});
        google.visualization.events.addListener(tree, 'onmouseover', function(ev) {
            $('mouseover_text').innerHTML = text_data[ev.row].text;
        });
        google.visualization.events.addListener(tree, 'select', function(ev) {
            if (text_data[tree.getSelection()[0].row].path == '<?php echo escape($path) ?>') {
                if (<?php echo ($level_min) ?> > 1) {
                    window.location.href='<?php echo $_SERVER['SCRIPT_NAME'] ?>?path=' + encodeURIComponent('<?php echo escape(dirname($path)) ?>') + '&level=<?php echo ($level_min-1) ?>';
                }
            } else if (text_data[tree.getSelection()[0].row].path != 'Files') {
                $('chart_div').innerHTML = 'Loading...';
                window.location.href='<?php echo $_SERVER['SCRIPT_NAME'] ?>?path=' + encodeURIComponent(text_data[tree.getSelection()[0].row].path) + '&level=<?php echo ($level_min+1) ?>';
                return false;
            }
        });
      }
      function $(el) {
        return document.getElementById(el);
      }
    </script>
  </head>
  <body style="font-family: Verdana, Arial">
    <center><small>Click a rectangle to dig deeper. Click the grey rectangle to go back up.</small></center><br/>
    <div id="chart_div" style="width: 100%; height: 90%;"></div>
    <div id="mouseover_text"></div>
  </body>
</html>
